import types
from yggdrasil.metaschema.datatypes.ClassMetaschemaType import (
    ClassMetaschemaType)


class FunctionMetaschemaType(ClassMetaschemaType):
    r"""Type for evaluating functions."""

    name = 'function'
    description = 'Type for callable Python functions.'
    python_types = (types.BuiltinFunctionType, types.FunctionType,
                    types.BuiltinMethodType, types.MethodType)

    @classmethod
    def _generate_data(cls, typedef):
        r"""Generate mock data for the specified type.

        Args:
            typedef (dict): Type definition.

        Returns:
            object: Python object of the specified type.

        """
        def example_func(x):  # pragma: debug
            return x
        return example_func



# https://www.python.org/dev/peps/pep-0508/
from pip._vendor.packaging.requirements import Requirement, InvalidRequirement
import os
import sys
import argparse
from ci_setup import call_conda_command


def prune(fname_in, fname_out=None, excl_suffix=None):
    r"""Prune a requirements.txt file to remove/select dependencies that are
    dependent on the current environment.

    Args:
        fname_in (str, list): Full path to one or more requirements files that
            should be read.
        fname_out (str, optional): Full path to requirements file that should be
            created. Defaults to None and is set to <fname_in[0]>_pruned.txt.
        excl_suffix (str, optional): Lines ending in this string will be
            excluded. Defaults to None and is ignored.

    Returns:
        str: Full path to created file.

    """
    if not isinstance(fname_in, (list, tuple)):
        fname_in = [fname_in]
    new_lines = []
    for ifname_in in fname_in:
        with open(ifname_in, 'r') as fd:
            old_lines = fd.readlines()
        for line in old_lines:
            line = line.strip()
            if line.startswith('#'):
                continue
            if excl_suffix and line.endswith(excl_suffix):
                continue
            try:
                req = Requirement(line.split('#')[0].strip())
                if req.marker and (not req.marker.evaluate()):
                    continue
                new_lines.append(req.name + str(req.specifier))
            except InvalidRequirement as e:
                print(e)
                continue
    # Write file
    if fname_out is None:
        fname_out = '_pruned'.join(os.path.splitext(fname_in[0]))
    with open(fname_out, 'w') as fd:
        fd.write('\n'.join(new_lines))
    return fname_out


def locate_conda_exe(conda_env, name):
    r"""Determine the full path to an executable in a specific conda environment.

    Args:
        conda_env (str): Name of conda environment that executable should be
            returned for.
        name (str): Name of the executable to locate.

    Returns:
        str: Full path to the executable.

    """
    conda_prefix = os.environ.get('CONDA_PREFIX', None)
    assert(conda_prefix)
    if os.path.dirname(conda_prefix).endswith('envs'):
        conda_prefix = os.path.dirname(conda_prefix)
    else:
        conda_prefix = os.path.join(conda_prefix, 'envs')
    if (sys.platform in ['win32', 'cygwin']):
        if not name.endswith('.exe'):
            name += '.exe'
        if name.startswith('python'):
            out = os.path.join(conda_prefix, conda_env, name)
        else:
            out = os.path.join(conda_prefix, conda_env,
                               'Scripts', name)
    else:
        out = os.path.join(conda_prefix, conda_env, 'bin', name)
    try:
        assert(os.path.isfile(out))
    except AssertionError:
        out = os.path.expanduser(os.path.join('~', '.conda', 'envs', name))
        if not os.path.isfile(out):
            raise
    return out


def install_from_requirements(method, fname_in, conda_env=None,
                              user=False):
    r"""Install packages via pip or conda from one or more pip-style
    requirements file(s).

    Args:
        method (str): Installation method; either 'pip' or 'conda'.
        fname_in (str, list): Full path to one or more requirements files that
            should be read.
        conda_env (str, optional): Name of conda environment that requirements
            should be installed into. Defaults to None and is ignored.
        user (bool, optional): If True, install in user mode. Defaults to
            False.

    """
    if method not in ['pip', 'conda']:
        raise ValueError("Invalid method: '%s'" % method)
    elif method == 'pip':
        excl_suffix = '# conda'
    elif method == 'conda':
        excl_suffix = '# pip'
    temp_file = prune(fname_in, excl_suffix=excl_suffix)
    try:
        if method == 'conda':
            args = ['conda', 'install', '-y']
            if conda_env:
                args += ['--name', conda_env]
            args += ['--file', temp_file]
            if user:
                args.append('--user')
        elif method == 'pip':
            if conda_env:
                pip_cmd = locate_conda_exe(conda_env, 'pip')
            else:
                pip_cmd = 'pip'
            args = [pip_cmd, 'install', '-r', temp_file]
            if user:
                args.append('--user')
        print(call_conda_command(args))
    except BaseException:
        if os.path.isfile(temp_file):
            with open(temp_file, 'r') as fd:
                print(fd.read())
        raise
    finally:
        if os.path.isfile(temp_file):
            os.remove(temp_file)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "Install dependencies via pip or conda from one or more "
        "pip-style requirements files.")
    parser.add_argument('method', choices=['conda', 'pip'],
                        help=("Method that should be used to install the "
                              "dependencies."))
    parser.add_argument('files', nargs='+',
                        help='One or more pip-style requirements files.')
    parser.add_argument('--conda-env', default=None,
                        help=('Conda environment that requirements should be '
                              'installed into.'))
    parser.add_argument('--user', action='store_true',
                        help=('Install in user mode.'))
    args = parser.parse_args()
    install_from_requirements(args.method, args.files, conda_env=args.conda_env,
                              user=args.user)



from yggdrasil.tests import YggTestClass


class TestFilterBase(YggTestClass):
    r"""Test for FilterBase communication flass."""

    filter = 'FilterBase'
    skip_comm_check = True
    
    @property
    def cls(self):
        r"""str: Communication class."""
        return self.filter

    @property
    def mod(self):
        r"""str: Absolute module import."""
        return 'yggdrasil.communication.filters.%s' % self.cls

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for tested class."""
        out = {}
        opt = self.get_options()
        if opt:
            out = opt[0].get('kwargs', {})
        return out

    def test_filter(self):
        r"""Test filter."""
        for x in self.get_options():
            inst = self.import_cls(**x.get('kwargs', {}))
            for msg in x.get('pass', []):
                self.assert_equal(inst(msg), True)
            for msg in x.get('fail', []):
                self.assert_equal(inst(msg), False)
            for msg, err in x.get('error', []):
                self.assert_raises(err, inst, msg)



import os
import sys
import copy
import pprint
import numpy as np
from scipy.io import netcdf
from yggdrasil import units
from yggdrasil.communication.FileComm import FileComm


class NetCDFFileComm(FileComm):
    r"""Class for handling I/O from/to an netCDF file.


    Args:
        read_attributes (bool, optional): If True, the attributes are read
            in as well as the variables. Defaults to False.
        variables (list, optional): List of variables to read in. If
            not provided, all variables will be read.
        version (int, optional): Version of netCDF format that should be
            used. Defaults to 1. Options are 1 (classic format) and
            2 (64-bit offset format).
        **kwargs: Additional keywords arguments are passed to parent class.

    """

    _filetype = 'netcdf'
    _schema_subtype_description = ('The file is read/written as netCDF.')
    _schema_properties = {
        'read_attributes': {'type': 'boolean', 'default': False},
        'variables': {'type': 'array', 'items': {'type': 'string'}},
        'version': {'type': 'integer', 'enum': [1, 2], 'default': 1}}
    _default_extension = '.nc'
    _mode_as_bytes = False
    _synchronous_read = True

    def __init__(self, *args, **kwargs):
        self._fd_netcdf = None
        kwargs['read_meth'] = 'read'
        self._last_size = 0
        return super(NetCDFFileComm, self).__init__(*args, **kwargs)

    @classmethod
    def get_testing_options(cls):
        r"""Method to return a dictionary of testing options for this class.

        Returns:
            dict: Dictionary of variables to use for testing. Key/value pairs:
                kwargs (dict): Keyword arguments for comms tested with the
                    provided content.
                send (list): List of objects to send to test file.
                recv (list): List of objects that will be received from a test
                    file that was sent the messages in 'send'.
                contents (bytes): Bytes contents of test file created by sending
                    the messages in 'send'.

        """
        data = {
            'time': units.add_units(np.arange(10).astype('float32'), 's'),
            'x': np.array(['a', 'hello', 'c'], 'S5')}
        data_add = {
            'space': units.add_units(np.ones((5, 5), 'int64'), 'mol')}
        out = {'kwargs': {},
               'exact_contents': False,
               'msg': data,
               'dict': False,
               'objects': [data, data_add],
               'send': [data, data_add],
               'recv': [dict(data, **data_add)],
               'recv_partial': [[data], [dict(data, **data_add)]],
               'contents': (
                   b'CDF\x01\x00\x00\x00\x00\x00\x00\x00\n\x00\x00'
                   b'\x00\x05\x00\x00\x00\x04time\x00\x00\x00\n\x00'
                   b'\x00\x00\x01x\x00\x00\x00\x00\x00\x00\x05\x00'
                   b'\x00\x00\x02x1\x00\x00\x00\x00\x00\x03\x00\x00'
                   b'\x00\x05space\x00\x00\x00\x00\x00\x00\x05\x00'
                   b'\x00\x00\x06space1\x00\x00\x00\x00\x00\x05\x00'
                   b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0b\x00'
                   b'\x00\x00\x03\x00\x00\x00\x04time\x00\x00\x00\x01'
                   b'\x00\x00\x00\x00\x00\x00\x00\x0c\x00\x00\x00\x01'
                   b'\x00\x00\x00\x05units\x00\x00\x00\x00\x00\x00'
                   b'\x02\x00\x00\x00\x01s\x00\x00\x00\x00\x00\x00'
                   b'\x05\x00\x00\x00(\x00\x00\x01\x0c\x00\x00\x00'
                   b'\x05space\x00\x00\x00\x00\x00\x00\x02\x00\x00'
                   b'\x00\x03\x00\x00\x00\x04\x00\x00\x00\x0c\x00\x00'
                   b'\x00\x01\x00\x00\x00\x05units\x00\x00\x00\x00'
                   b'\x00\x00\x02\x00\x00\x00\x03mol\x00\x00\x00\x00'
                   b'\x04\x00\x00\x00d\x00\x00\x014\x00\x00\x00\x01x'
                   b'\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00'
                   b'\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                   b'\x00\x00\x02\x00\x00\x00\x10\x00\x00\x01\x98\x00'
                   b'\x00\x00\x00?\x80\x00\x00@\x00\x00\x00@@\x00\x00'
                   b'@\x80\x00\x00@\xa0\x00\x00@\xc0\x00\x00@\xe0\x00'
                   b'\x00A\x00\x00\x00A\x10\x00\x00\x00\x00\x00\x01'
                   b'\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01'
                   b'\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01'
                   b'\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01'
                   b'\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01'
                   b'\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01'
                   b'\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01'
                   b'\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01'
                   b'\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x01'
                   b'ahc\x00e\x00\x00l\x00\x00l\x00\x00o\x00\x00')}
        return out

    @property
    def concats_as_str(self):
        r"""bool: True if concatenating file contents result in a
        valid file."""
        return False

    def serialize(self, obj, **kwargs):
        r"""Don't serialize for netCDF since using a serializer
        is inefficient."""
        return obj

    def deserialize(self, msg, **kwargs):
        r"""Don't deserialize for netCDF since using a serializer
        is inefficient."""
        return msg, {}

    def transform_type_recv(self, x):
        x_dtype = np.dtype(x.dtype)
        typecode, size = x_dtype.char, x_dtype.itemsize
        if (typecode == 'c') and (x.ndim > 1):
            size = x.shape[0]
            new_shape = x.shape[1:]
            x_str = np.zeros(new_shape, 'S%d' % size)
            for index in np.ndindex(*new_shape):
                for i in range(size):
                    x_str[index] += x[tuple([i, *index])]
            x = x_str
        else:
            x = x.copy()
            if ((((sys.byteorder == 'little') and (x.dtype.byteorder == '>'))
                 or ((sys.byteorder == 'big') and (x.dtype.byteorder == '<')))):
                x = x.astype(x.dtype.name)
        return x

    def transform_type_send(self, x):
        x_dtype = np.dtype(x.dtype)
        typecode, size = x_dtype.char, x_dtype.itemsize
        typecode_map = {'l': 'i', 'q': 'i'}
        if (typecode, size) not in netcdf.REVERSE:
            REVERSE_keys = list(netcdf.REVERSE.keys())
            REVERSE_typecode = [k[0] for k in REVERSE_keys]
            typecode = typecode_map.get(typecode, typecode)
            if typecode == 'S':
                x_str = np.zeros(tuple([size] + list(x.shape)), 'S1')
                for index in np.ndindex(*x.shape):
                    for i in range(size):
                        if i == len(x[index]):
                            x_str[tuple([i, *index])] = '\0'
                        elif i < len(x[index]):
                            x_str[tuple([i, *index])] = x[index][i:(i + 1)]
                x = x_str
            elif typecode in REVERSE_typecode:
                x = x.astype(np.dtype(*REVERSE_keys[
                    REVERSE_typecode.index(typecode)]))
            else:  # pragma: debug
                raise RuntimeError(
                    ("Type (%s, %d) is not in set accepted by "
                     "netCDF %s.")
                    % (typecode, size, pprint.pformat(netcdf.REVERSE)))
        return x
    
    # Methods related to position in the file/series
    @property
    def file_size(self):
        r"""int: Current size of file."""
        out = 0
        if os.path.isfile(self.current_address):
            out = os.stat(self.current_address).st_size
        return out

    def file_tell(self):
        r"""int: Current position in the file."""
        return self._last_size
    
    def file_seek(self, pos, whence=os.SEEK_SET):
        r"""Move in the file to the specified position.

        Args:
            pos (int): Position (in bytes) to move file to.
            whence (int, optional): Flag indicating position that pos
                is relative to. 0 for the beginning of the file, 1 for
                from the current location, and 2 from the end of the
                file.

        """
        super(NetCDFFileComm, self).file_seek(pos, whence)
        if whence == 0:
            self._last_size = pos
        elif whence == 1:  # pragma: no cover
            self._last_size = min(self.file_size, self._last_size + pos)
        elif whence == 2:
            self._last_size = self.file_size
        
    def file_flush(self):
        r"""Flush the file."""
        super(NetCDFFileComm, self).file_flush()
        if self._fd_netcdf is not None:
            self._fd_netcdf.flush()
            self._fd_netcdf.sync()
        
    def _file_open(self, address, mode):
        self._last_size = 0
        if ((((not os.path.isfile(address)) or (os.stat(address).st_size == 0))
             and (mode == 'r'))):
            # NetCDF dosn't allow opening an empty file for read
            # because it will not contain the opening version bytes
            return super(NetCDFFileComm, self)._file_open(address, mode)
        self._fd_netcdf = netcdf.netcdf_file(address, mode,
                                             mmap=True,
                                             version=self.version)
        return self._fd_netcdf.fp

    def _file_close(self):
        super(NetCDFFileComm, self)._file_close()
        if self._fd_netcdf is not None:
            self._fd_netcdf.close()
            self._fd_netcdf = None

    def _file_refresh(self):
        prev_pos = self.file_tell()
        self._file_close()
        self._fd = self._file_open(self.current_address,
                                   self.open_mode)
        self.file_seek(prev_pos)

    def _file_send(self, msg):
        assert(isinstance(msg, dict))
        for k, v in msg.items():
            if isinstance(v, np.ndarray):
                dims = []
                v = self.transform_type_send(v)
                for i, d in enumerate(v.shape):
                    if i == 0:
                        idim = k
                    else:
                        idim = '%s%d' % (k, i)
                    dims.append(idim)
                    self._fd_netcdf.createDimension(idim, d)
                var = self._fd_netcdf.createVariable(k, v.dtype, dims)
                var[:] = v
                if units.has_units(v):
                    var.units = units.get_units(v)
            else:  # pragma: debug
                raise TypeError("Type '%s' no supported." % type(msg))
        self._last_size = self.file_size
        self.fd.seek(self._last_size)

    @property
    def remaining_bytes(self):
        r"""int: Remaining bytes in the file."""
        if self.is_open and ((self._fd_netcdf is None) or self.append):
            self._file_refresh()
        return super(NetCDFFileComm, self).remaining_bytes
        
    def _file_recv(self):
        out = {}
        if self.is_open and ((self._fd_netcdf is None) or self.append):
            self._file_refresh()
        if self.file_size > self._last_size:
            variables = self.variables
            if not variables:
                variables = list(self._fd_netcdf.variables.keys())
            for v in variables:
                out[v] = self.transform_type_recv(
                    self._fd_netcdf.variables[v][:])
                if hasattr(self._fd_netcdf.variables[v], 'units'):
                    out[v] = units.add_units(
                        out[v], self._fd_netcdf.variables[v].units)
            self._last_size = self.file_size
            self.fd.seek(self._last_size)
        return copy.deepcopy(out)



import sys
import numpy as np
from matplotlib import pyplot as plt
from yggdrasil import units
from yggdrasil.communication.AsciiTableComm import AsciiTableComm


def main(fileA, fileB, example_name):
    r"""Method to plot comparison of two results."""
    if example_name == 'timesync2':
        figsize = (10.8, 4.8)
        fig = plt.figure(figsize=figsize)
        ax1 = fig.add_subplot(2, 2, 1)
        ax2 = fig.add_subplot(2, 2, 2, sharey=ax1)
        ax3 = fig.add_subplot(2, 2, 3, sharex=ax1)
    else:
        figsize = None
        fig = plt.figure(figsize=figsize)
        ax1 = fig.add_subplot(1, 1, 1)
    dataA = AsciiTableComm('test', address=fileA, direction='recv',
                           as_array=True).recv_dict()[1]
    dataB = AsciiTableComm('test', address=fileB, direction='recv',
                           as_array=True).recv_dict()[1]
    xtrue = np.sin(2.0 * np.pi * dataA['time'] / units.add_units(10, 'day'))
    ytrue = np.cos(2.0 * np.pi * dataA['time'] / units.add_units(5, 'day'))
    ztrue = -np.cos(2.0 * np.pi * dataA['time'] / units.add_units(20, 'day'))
    atrue = np.sin(2.0 * np.pi * dataA['time'] / units.add_units(2.5, 'day'))
    btrue = np.cos(2.0 * np.pi * dataA['time'] / units.add_units(2.5, 'day'))
    if example_name == 'timesync2':
        ytrue *= 2.0
        dataB['x'] = 2.0 * dataB['xvar']
        dataB['y'] = dataB['yvar']
    ax1.plot(dataA['time'], xtrue, 'k')
    ax1.plot(dataA['time'], ytrue, 'k')
    ax1.plot(dataA['time'], dataA['x'], 'b-', label='x (model A)')
    ax1.plot(dataA['time'], dataA['y'], 'r-', label='y (model A)')
    ax1.plot(dataB['time'].to(dataA['time'].units), dataB['x'],
             'b--', label='x (model B)')
    ax1.plot(dataB['time'].to(dataA['time'].units), dataB['y'],
             'r--', label='y (model B)')
    ax1.set_xlabel('Time (%s)' % dataA['time'].units)
    ax1.set_ylabel('State Value')
    ax1.legend(loc='upper center', bbox_to_anchor=(0.5, 1.15), ncol=2)
    if example_name == 'timesync2':
        ax2.plot(dataA['time'], ztrue, 'k')
        ax2.plot(dataA['time'], dataA['z1'], 'o-', label='z1 (model A)')
        ax2.plot(dataA['time'], dataA['z2'], 'o-', label='z2 (model A)')
        ax2.plot(dataB['time'].to(dataA['time'].units), dataB['z'],
                 'o--', label='z (model B)')
        ax2.legend(loc='upper center', bbox_to_anchor=(0.5, 1.15), ncol=2)
        ax3.plot(dataA['time'], atrue, 'k')
        ax3.plot(dataA['time'], btrue, 'k')
        ax3.plot(dataA['time'], dataA['a'], 'c-', label='a (model A)')
        ax3.plot(dataA['time'], dataA['b'], 'm-', label='b (model A)')
        ax3.plot(dataB['time'].to(dataA['time'].units), dataB['a'],
                 'c--', label='a (model B)')
        ax3.plot(dataB['time'].to(dataA['time'].units), dataB['b'],
                 'm--', label='b (model B)')
        ax3.legend(loc='upper center', bbox_to_anchor=(0.5, 1.15), ncol=2)
        plt.subplots_adjust(left=0.075, right=0.95, wspace=0.1, hspace=0.45)
    plt.show()


if __name__ == '__main__':
    # Take time step from the first argument
    if len(sys.argv) >= 4:
        example_name = sys.argv[3]
    else:
        example_name = 'timesync1'
    main(sys.argv[1], sys.argv[2], example_name)



import os
import git
import copy
import shutil
import subprocess
import tempfile
import logging
import pystache
import io as sio
import warnings
import xml.etree.ElementTree as ET
from yggdrasil import tools, platform
from yggdrasil.components import import_component
from yggdrasil.drivers.ExecutableModelDriver import ExecutableModelDriver
from yggdrasil.drivers.CPPModelDriver import CPPModelDriver


logger = logging.getLogger(__name__)


class OSRModelDriver(ExecutableModelDriver):
    r"""Class for running OpenSimRoot model.

    Args:
        sync_vars_in (list, optional): Variables that should be synchronized
            from other models. Defaults to [].
        sync_vars_out (list, optional): Variables that should be synchronized
            to other models. Defaults to [].
        copy_xml_to_osr (bool, optional): If True, the XML file(s) will
            be copied to the OSR repository InputFiles direcitory before
            running. This is necessary if the XML file(s) use any of the
            files located there since OSR always assumes the included
            file paths are relative. Defaults to False.
        update_interval (float, optional): Max simulation interval at which
            synchronization should occur (in days). Defaults to 1.0 if not
            provided. If the XML input file loads additional export modules
            that output at a shorter rate, the existing table of values will
            be extrapolated.

    """
    _schema_subtype_description = 'Model is an OSR model.'
    _schema_properties = {
        'sync_vars_in': {'type': 'array', 'items': {'type': 'string'},
                         'default': []},
        'sync_vars_out': {'type': 'array', 'items': {'type': 'string'},
                          'default': []},
        'copy_xml_to_osr': {'type': 'boolean', 'default': False},
        'update_interval': {'type': 'object',
                            'additionalProperties': {'type': 'float'},
                            'default': {'timesync': 1.0}}}
    language = 'osr'
    language_ext = '.xml'
    base_languages = ['cpp']
    repository = None
    executable_path = None
    repository_url = "https://gitlab.com/langmm/OpenSimRoot.git"
    repository_branch = "volatile_active"
    _config_keys = ['repository']
    _config_attr_map = [{'attr': 'repository',
                         'key': 'repository'}]

    @staticmethod
    def after_registration(cls, **kwargs):
        r"""Operations that should be performed to modify class attributes after
        registration.
        """
        ExecutableModelDriver.after_registration(cls, **kwargs)
        if cls.repository is not None:
            if platform._is_win:  # pragma: windows
                cls.executable_path = os.path.join(
                    cls.repository, 'OpenSimRoot',
                    'StaticBuild_win64', 'OpenSimRootYgg') + '.exe'
            else:
                cls.executable_path = os.path.join(
                    cls.repository, 'OpenSimRoot',
                    'StaticBuild', 'OpenSimRootYgg')
        
    def parse_arguments(self, *args, **kwargs):
        r"""Sort model arguments to determine which one is the executable
        and which ones are arguments.

        Args:
            *args: Additional arguments are passed to the parent class's method.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        """
        super(OSRModelDriver, self).parse_arguments(*args, **kwargs)
        self.model_file_orig = self.model_file
        self.model_file = '_copy'.join(os.path.splitext(self.model_file_orig))
        if self.copy_xml_to_osr and (self.repository is not None):
            self.model_file = os.path.join(
                self.repository, 'OpenSimRoot', 'InputFiles',
                os.path.basename(self.model_file))
        # if not (isinstance(self.executable_path, str)
        #         and os.path.isfile(self.executable_path)):
        self.compile_osr()
        assert(os.path.isfile(self.executable_path))

    @classmethod
    def compile_osr(cls, target='OpenSimRootYgg'):
        r"""Compile the OpenSimRoot executable with the yggdrasil flag set.

        Args:
            target (str): Make target that should be build. Defaults to
                'OpenSimRootYgg' (the yggdrasil-instrumented version of the
                OSR executable).

        """
        if (cls.repository is not None) and CPPModelDriver.is_installed():
            toolname = None
            # toolname = CPPModelDriver.get_tool('compiler',
            #                                    return_prop='name',
            #                                    default=None)
            cwd = os.path.join(cls.repository, 'OpenSimRoot')
            flags = ['-j4']
            env = copy.deepcopy(os.environ)
            if platform._is_win:  # pragma: windows
                toolname = 'cl'
                env['YGG_OSR_TOOL'] = toolname
                if toolname == 'cl':
                    cl_path = shutil.which(toolname + '.exe')
                    print('cl.exe', shutil.which(toolname), cl_path)
                    subprocess.check_call(['yggccflags', '--cpp',
                                           '--toolname=%s' % toolname])
                    subprocess.check_call(['yggldflags', '--cpp',
                                           '--toolname=%s' % toolname])
                    if cl_path:
                        msvc_bin = os.path.dirname(cl_path)
                        env['YGG_OSR_CXX'] = cl_path
                        env['YGG_OSR_LINK'] = os.path.join(msvc_bin, 'link.exe')
                        for k in ['CL', '_CL_']:
                            v = os.environ.get(k, None)
                            if v is not None:
                                env[k] = v.replace('/', '-').replace('\\', '/')
                    else:  # pragma: debug
                        env.pop('YGG_OSR_TOOL')
                        warnings.warn(
                            "The MSVC compiler is not installed. Be aware "
                            "that the GNU compiler takes a *very* long time "
                            "to compile OpenSimRoot against yggdrasil on "
                            "Windows (> 1 hr).")
                cwd = os.path.join(cwd, 'StaticBuild_win64')
            else:
                cwd = os.path.join(cwd, 'StaticBuild')
            if target != 'cleanygg':
                for x in cls.base_languages:
                    base_cls = import_component('model', x)
                    base_cls.compile_dependencies(toolname=toolname)
            elif not os.path.isfile(cls.executable_path):
                return
            cmd = ['make', target] + flags
            subprocess.check_call(cmd, cwd=cwd, env=env)

    def write_wrappers(self, **kwargs):
        r"""Write any wrappers needed to compile and/or run a model.

        Args:
            **kwargs: Keyword arguments are passed to the parent class's
                method.

        Returns:
            list: Full paths to any created wrappers.

        """
        out = super(OSRModelDriver, self).write_wrappers(**kwargs)
        out.append(self.wrap_xml(self.model_file_orig, self.model_file))
        return out

    def wrap_xml(self, src, dst):
        r"""Wrap the input XML, adding tags for the inputs/outputs specified for
        this model.

        Args:
            src (str): Full path to the XML file that should be wrapped.
            dst (str): Full path to the location where the updated XML should
                be saved.

        Returns:
            str: Full path to XML file produced.

        """
        with open(src, 'r') as fd:
            src_contents = fd.read()
        src_contents = pystache.render(
            sio.StringIO(src_contents).getvalue(), self.set_env())
        root = ET.fromstring(src_contents)
        timesync = self.timesync
        assert(timesync)
        if not isinstance(timesync, list):
            timesync = [timesync]
        for tsync in reversed(timesync):
            ivars = tsync.get('inputs', [])
            ovars = tsync.get('outputs', [])
            assert(isinstance(ivars, list))
            assert(isinstance(ovars, list))
            ivars = ' '.join(ivars)
            ovars = ' '.join(ovars)
            tupdate = self.update_interval.get(tsync['name'], 1.0)
            cask = ET.Element('SimulaCask',
                              attrib={'name': tsync['name'],
                                      'interval': str(tupdate)})
            if ivars:
                icask = ET.Element('SimulaCaskInputs')
                icask.text = ivars
                cask.insert(-1, icask)
            if ovars:
                ocask = ET.Element('SimulaCaskOutputs')
                ocask.text = ovars
                cask.insert(-1, ocask)
            probe = ET.Element("SimulaBase", name="probeCaskObjects")
            probe_vals = [ET.Element("SimulaConstant",
                                     name="run", type="bool"),
                          ET.Element("SimulaConstant",
                                     name="timeInterval", type="time"),
                          ET.Element("SimulaConstant",
                                     name="requestedVariables",
                                     type="string")]
            probe_vals[0].text = '1'
            probe_vals[1].text = str(tupdate)
            probe_vals[2].text = ', '.join(ivars.split())
            probe.extend(probe_vals)
            probe_directive = ET.Element(
                'SimulaDirective',
                attrib={'path': "/simulationControls/outputParameters"})
            probe_directive.append(probe)
            root.extend([cask, probe_directive])
        tree = ET.ElementTree(root)
        tree.write(dst)
        return dst

    @classmethod
    def is_language_installed(cls):
        r"""Determine if the interpreter/compiler for the associated programming
        language is installed.

        Returns:
            bool: True if the language interpreter/compiler is installed.

        """
        return (cls.repository is not None)
        
    @classmethod
    def configure_executable_type(cls, cfg):
        r"""Add configuration options specific in the executable type
        before the libraries are configured.

        Args:
            cfg (CisConfigParser): Config class that options should be set for.
        
        Returns:
            list: Section, option, description tuples for options that could not
                be set.

        """
        # TODO: Move clone + compile to install
        out = super(OSRModelDriver, cls).configure_executable_type(cfg)
        opt = 'repository'
        desc = 'The full path to the OpenSimRoot repository.'
        # if platform._is_win:  # pragma: windows
        #     out.append((cls.language, opt, desc))
        #     return out
        if not cfg.has_option(cls.language, opt):
            fname = 'OpenSimRoot'
            fpath = tools.locate_file(fname)
            if not fpath:
                logger.info('Could not locate %s, attempting to clone' % fname)
                try:
                    fpath = os.path.join(tempfile.gettempdir(), fname)
                    if not os.path.isdir(fpath):  # pragma: config
                        git.Repo.clone_from(cls.repository_url, fpath,
                                            branch=cls.repository_branch)
                except BaseException as e:  # pragma: debug
                    logger.info('Failed to clone from %s. error = %s'
                                % (cls.repository_url, str(e)))
                    out.append((cls.language, opt, desc))
            logger.info('Located %s: %s' % (fname, fpath))
            cfg.set(cls.language, opt, fpath)
        return out
        
    @classmethod
    def set_env_class(cls, **kwargs):
        r"""Set environment variables that are instance independent.

        Args:
            **kwargs: Additional keyword arguments are passed to the
                parent class's method.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(OSRModelDriver, cls).set_env_class(**kwargs)
        if cls.repository is not None:
            out['OSR_REPOSITORY'] = cls.repository
        kwargs['existing'] = out
        out = CPPModelDriver.set_env_class(**kwargs)
        return out
        
    @classmethod
    def language_executable(cls, **kwargs):
        r"""Command required to compile/run a model written in this language
        from the command line.

        Returns:
            str: Name of (or path to) compiler/interpreter executable required
                to run the compiler/interpreter from the command line.

        """
        return cls.executable_path
        
    @classmethod
    def executable_command(cls, args, **kwargs):
        r"""Compose a command for running a program using the exectuable for
        this language (compiler/interpreter) with the provided arguments.

        Args:
            args (list): The program that returned command should run and any
                arguments that should be provided to it.
            unused_kwargs (dict, optional): Existing dictionary that unused
                keyword arguments should be added to. Defaults to None and is
                ignored.
            **kwargs: Additional keyword arguments are passed to the parent
                class.

        Returns:
            list: Arguments composing the command required to run the program
                from the command line using the executable for this language.

        """
        args = [cls.language_executable()] + args
        return super(OSRModelDriver, cls).executable_command(args, **kwargs)

    @classmethod
    def cleanup_dependencies(cls, *args, **kwargs):
        r"""Cleanup dependencies."""
        cls.compile_osr(target='cleanygg')
        super(OSRModelDriver, cls).cleanup_dependencies(*args, **kwargs)



import os
import trimesh
import argparse
import numpy as np
from yggdrasil import units
from yggdrasil.languages.Python.YggInterface import YggRpcClient, YggOutput


def run(mesh, tmin, tmax, tstep):
    mass = 2000.0
    mesh_out = YggOutput('mesh')
    mass = units.add_units(mass, 'g')
    tmin = units.add_units(tmin, 'hrs')
    tmax = units.add_units(tmax, 'hrs')
    tstep = units.add_units(tstep, 'hrs')
    t = tmin

    while t <= tmax:

        # Grow mesh
        # (pretend this is a biologically complex calculation)
        scale = units.get_data(mass * t / units.add_units(1.5e6, 'g*hrs'))
        mesh.vertices[:, 2] += mesh.vertices[:, 2] * scale

        # Send mesh to output for this timestep
        flag = mesh_out.send(mesh)
        if not flag:
            raise Exception("Error sending mesh to output")

        # Advance time step
        t += tstep

    return mesh


# Parse command-line arguments
if __name__ == "__main__":
    parser = argparse.ArgumentParser("Simulate plant growth over time.")
    parser.add_argument('tmin', help='Starting time (in hours)', type=float)
    parser.add_argument('tmax', help='Ending time (in hours)', type=float)
    parser.add_argument('tstep', help='Time step (in hours)', type=float)
    parser.add_argument('--meshfile', help='Path to file where mesh is stored.',
                        default='../meshes/plants-2.obj')
    args = parser.parse_args()
    mesh = trimesh.load_mesh(args.meshfile)
    run(mesh, args.tmin, args.tmax, args.tstep)



import os
import glob
import pickle
import inspect
import trimesh
import difflib
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.lexers.data import YamlLexer
from pygments.formatters import Terminal256Formatter
from pygments.lexers.c_cpp import CLexer, CppLexer
from pygments.lexers.r import SLexer
from pygments.lexers.fortran import FortranLexer
import matplotlib.pyplot as plt


ext_map = {'.py': 'python',
           '.c': 'c',
           '.cpp': 'c++',
           '.f90': 'fortran',
           '.R': 'R',
           '.yml': 'yaml'}
lexer_map = {'python': PythonLexer,
             'yaml': YamlLexer,
             'c': CLexer,
             'c++': CppLexer,
             'R': SLexer,
             'fortran': FortranLexer}


def number_lines(lines, for_diff=False):
    as_str = False
    if isinstance(lines, str):
        as_str = True
        lines = lines.splitlines()
    out = []
    i = 0
    for line in lines:
        if for_diff and line.startswith(('-', '?')):
            out.append('    ' + line)
        else:
            i += 1
            out.append('%2d: %s' % (i, line))
    if as_str:
        return '\n'.join(out)
    return out


def print_lines_numbered(lines, end=None, for_diff=False):
    if isinstance(lines, str):
        lines = lines.splitlines()
    lines_no = number_lines(lines, for_diff=for_diff)
    print('\n'.join(lines_no), end=end)


def print_source_diff(fname1, fname2, language=None):
    src1 = print_source(fname1, language=language,
                        return_lines=True).splitlines()
    src2 = print_source(fname2, language=language,
                        return_lines=True).splitlines()
    diff = difflib.ndiff(src1, src2)
    print_lines_numbered(diff, for_diff=True)


def print_python_source_diff(x1, x2):
    src1 = print_python_source(x1, return_lines=True).splitlines()
    src2 = print_python_source(x2, return_lines=True).splitlines()
    diff = difflib.ndiff(src1, src2)
    print_lines_numbered(diff, for_diff=True)


def print_yaml_diff(fname1, fname2):
    src1 = print_yaml(fname1, return_lines=True).splitlines()
    src2 = print_yaml(fname2, return_lines=True).splitlines()
    diff = difflib.ndiff(src1, src2)
    print_lines_numbered(diff, for_diff=True)


def print_source(fname, language=None, return_lines=False):
    with open(fname, 'r') as fd:
        lines = fd.read()
    if language is None:
        language = ext_map[os.path.splitext(fname)[-1]]
    out = highlight(lines, lexer_map[language](), Terminal256Formatter())
    if return_lines:
        return out
    print_lines_numbered(out)


def print_python_source(x, return_lines=False):
    out = highlight(inspect.getsource(x), PythonLexer(),
                    Terminal256Formatter())
    if return_lines:
        return out
    print_lines_numbered(out)


def print_yaml(fname, return_lines=False, no_numbers=False):
    with open(fname, 'r') as fd:
        lines = fd.read()
    out = highlight(lines, YamlLexer(), Terminal256Formatter())
    if return_lines:
        return out
    if no_numbers:
        print(out)
    else:
        print_lines_numbered(out)


def display_last_timestep(with_light=False, with_masses=False):
    last_mesh = sorted(glob.glob('output/mesh_*.obj'))[-1]
    mesh = trimesh.load_mesh(last_mesh)
    if with_light:
        last_light = sorted(glob.glob('output/light_*.pkl'))[-1]
        with open(last_light, 'rb') as fd:
            light = pickle.load(fd)
        mesh.visual.vertex_colors = trimesh.visual.interpolate(
            light/max(light))
    if with_masses:
        plot_mass()
        return mesh.show()
    return mesh.show()


def plot_mass():
    from yggdrasil.communication.AsciiTableComm import AsciiTableComm
    fd = AsciiTableComm('mass', address='output/mass.txt', direction='recv',
                        as_array=True)
    flag, masses = fd.recv_dict()
    plt.plot(masses['time'], masses['root_mass'], label='root mass')
    plt.plot(masses['time'],
             masses['plant_mass'].to(masses['root_mass'].units),
             label='plant_mass')
    plt.plot(masses['time'],
             masses['total_mass'].to(masses['root_mass'].units),
             label='total_mass')
    plt.xlabel('time (%s)' % masses['time'].units)
    plt.ylabel('mass (%s)' % masses['root_mass'].units)
    plt.legend()
    return plt.show()



import os
import copy
from yggdrasil import platform
from yggdrasil.drivers.CModelDriver import (
    CCompilerBase, CModelDriver, GCCCompiler, ClangCompiler,
    ClangLinker)


class CPPCompilerBase(CCompilerBase):
    r"""Base class for C++ compilers."""
    languages = ['c++']
    default_executable_env = 'CXX'
    default_flags_env = 'CXXFLAGS'
    cpp_std = 'c++14'
    search_path_flags = ['-E', '-v', '-xc++', '/dev/null']
    default_linker = None
    default_executable = None

    @classmethod
    def get_flags(cls, skip_standard_flag=False, **kwargs):
        r"""Get a list of compiler flags.

        Args:
            skip_standard_flag (bool, optional): If True, the C++ standard flag
                will not be added. Defaults to False.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Compiler flags.

        """
        out = super(CCompilerBase, cls).get_flags(**kwargs)
        # Add standard library flag
        if not skip_standard_flag:
            std_flag = None
            for i, a in enumerate(out):
                if a.startswith('-std='):
                    std_flag = i
                    break
            if std_flag is None:
                out.append('-std=%s' % cls.cpp_std)
        return out
    

class GPPCompiler(CPPCompilerBase, GCCCompiler):
    r"""Interface class for G++ compiler/linker."""
    toolname = 'g++'
    aliases = ['gnu-c++']


class ClangPPCompiler(CPPCompilerBase, ClangCompiler):
    r"""Interface class for clang++ compiler."""
    toolname = 'clang++'
    default_linker = 'clang++'
    # Set to False since ClangLinker has its own class to handle
    # conflict between versions of clang and ld.
    is_linker = False

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        if platform._is_win:  # pragma: windows
            cls.default_executable = 'clang'
        CPPCompilerBase.before_registration(cls)

    @classmethod
    def get_executable_command(cls, args, skip_flags=False, unused_kwargs=None,
                               **kwargs):
        r"""Determine the command required to run the tool using the specified
        arguments and options.

        Args:
            args (list): The arguments that should be passed to the tool. If
                skip_flags is False, these are treated as input files that will
                be used by the tool.
            **kwargs: Additional keyword arguments will be passed to the parent
                class's method.

        Returns:
            str: Output to stdout from the command execution.

        """
        if platform._is_win:  # pragma: windows
            for a in args:
                if a.endswith('.c'):
                    kwargs['skip_standard_flag'] = True
                    break
        return super(ClangPPCompiler, cls).get_executable_command(args, **kwargs)


class ClangPPLinker(ClangLinker):
    r"""Interface class for clang++ linker (calls to ld)."""
    toolname = ClangPPCompiler.toolname
    languages = ClangPPCompiler.languages
    default_executable = ClangPPCompiler.default_executable
    default_executable_env = ClangPPCompiler.default_executable_env
    toolset = ClangPPCompiler.toolset


class CPPModelDriver(CModelDriver):
    r"""Class for running C++ models."""
                
    _schema_subtype_description = ('Model is written in C++.')
    language = 'c++'
    language_ext = ['.cpp', '.CPP', '.cxx', '.C', '.c++', '.cc', '.cp', '.tcc',
                    '.hpp', '.HPP', '.hxx', '.H', '.h++', '.hh', '.hp', '.h']
    language_aliases = ['cpp']
    base_languages = ['c']
    interface_library = 'ygg++'
    # To prevent inheritance
    default_compiler = None
    default_linker = None
    function_param = dict(
        CModelDriver.function_param,
        input='YggInput {channel}(\"{channel_name}\", {channel_type});',
        output='YggOutput {channel}(\"{channel_name}\", {channel_type});',
        recv_heap='{channel}.recvRealloc',
        recv_stack='{channel}.recv',
        recv_function='{channel}.recvRealloc',
        send_function='{channel}.send',
        exec_prefix=('#include <iostream>\n'
                     '#include <exception>\n'),
        print_generic='std::cout << {object} << std::endl << std::flush;',
        error='throw \"{error_msg}\";',
        try_begin='try {',
        try_error_type='const std::exception&',
        try_except='}} catch ({error_type} {error_var}) {{',
        function_def_regex=(
            r'(?P<flag_type>.+?)\s*{function_name}\s*'
            r'\((?P<inputs>(?:[^{{\&])*?)'
            r'(?:,\s*(?P<outputs>'
            r'(?:\s*(?:[^\s\&]+)'
            r'(?:(?:\&\s+)|(?:\s+(?:\()?\&))'
            r'(?:[^{{])+)+))?\)\s*\{{'
            r'(?P<body>(?:.*?\n?)*?)'
            r'(?:(?:return +(?P<flag_var>.+?)?;(?:.*?\n?)*?\}})'
            r'|(?:\}}))'),
        outputs_def_regex=(
            r'\s*(?P<native_type>(?:[^\s])+)(\s+)?'
            r'(\()?(?P<ref>\&)(?(1)(?:\s*)|(?:\s+))'
            r'(?P<name>.+?)(?(2)(?:\)|(?:)))(?P<shape>(?:\[.+?\])+)?\s*(?:,|$)(?:\n)?'))
    include_arg_count = True
    include_channel_obj = False
    dont_declare_channel = True
    
    @staticmethod
    def after_registration(cls, **kwargs):
        r"""Operations that should be performed to modify class attributes after
        registration."""
        if cls.default_compiler is None:
            if platform._is_linux:
                cls.default_compiler = 'g++'
            elif platform._is_mac:
                cls.default_compiler = 'clang++'
            elif platform._is_win:  # pragma: windows
                cls.default_compiler = 'cl'
        cls.function_param['print'] = 'std::cout << "{message}" << std::endl;'
        CModelDriver.after_registration(cls, **kwargs)
        if kwargs.get('second_pass', False):
            return
        internal_libs = copy.deepcopy(cls.internal_libraries)
        internal_libs[cls.interface_library] = internal_libs.pop(
            CModelDriver.interface_library)
        internal_libs[cls.interface_library]['source'] = os.path.join(
            cls.get_language_dir(),
            os.path.splitext(os.path.basename(
                internal_libs[cls.interface_library]['source']))[0]
            + cls.language_ext[0])
        internal_libs[cls.interface_library]['include_dirs'].append(
            cls.get_language_dir())
        internal_libs[cls.interface_library]['language'] = cls.language
        cls.internal_libraries = internal_libs

    @classmethod
    def set_env_class(cls, **kwargs):
        r"""Set environment variables that are instance independent.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method and update_ld_library_path.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(CPPModelDriver, cls).set_env_class(**kwargs)
        out = CModelDriver.update_ld_library_path(out, **kwargs)
        return out
        
    @classmethod
    def write_try_except(cls, try_contents, except_contents, error_var='e',
                         error_type=None, **kwargs):
        r"""Return the lines required to complete a try/except block.

        Args:
            try_contents (list): Lines of code that should be executed inside
                the try block.
            except_contents (list): Lines of code that should be executed inside
                the except block.
            error_var (str, optional): Name of variable where the caught error
                should be stored. Defaults to 'e'. If '...', the catch clause
                will catch all errors, but there will not be a name error.
            error_type (str, optional): Name of error type that should be caught.
                If not provided, defaults to None and will be set based on the
                class function_param entry for 'try_error_type'. If '...', the
                catch clause will catch all errors and error_var will be
                ignored.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            Lines of code perfoming a try/except block.

        """
        if (error_type == '...') or (error_var == '...'):
            error_type = ''
            error_var = '...'
        kwargs.update(error_var=error_var, error_type=error_type)
        return super(CPPModelDriver, cls).write_try_except(
            try_contents, except_contents, **kwargs)

    @classmethod
    def output2input(cls, var, in_definition=True):
        r"""Perform conversion necessary to turn an output variable
        into an corresponding input that can be used to format a
        function definition.

        Args:
            var (dict): Variable definition.
            in_definition (bool, optional): If True, the returned
                dictionary corresponds to an input variable in a
                function definition. If False, the returned value
                will correspond to an input to a function. Defaults to
                True.

        Returns:
            dict: Updated variable definition.

        """
        out = super(CModelDriver, cls).output2input(var)
        if isinstance(var, dict):
            if in_definition:
                out = dict(out, name='&' + out['name'])
                if ((('shape' in out.get('datatype', {}))
                     or ('length' in out.get('datatype', {})))):
                    out['name'] = '(%s)' % out['name']
            else:
                if not (out.get('ref', False)
                        or out.get('is_length_var', False)):
                    out = dict(out, name='&' + out['name'])
        return out



import weakref
from yggdrasil.metaschema.properties.tests import (
    test_MetaschemaProperty as parent)


class Dummy(object):  # pragma: debug
    pass


class ValidArgsClass1(object):
    test_args = tuple([int(0), Dummy])
    test_kwargs = dict(c=int(1), d=Dummy)
    valid_args = [{'type': 'int'}, {'type': 'class'}]
    valid_kwargs = {'c': {'type': 'int'},
                    'd': {'type': 'class'}}
    invalid_args = [{'type': 'int'}, {'type': 'float'}]
    invalid_kwargs = {'c': {'type': 'float'},
                      'd': {'type': 'float'}}

    def __init__(self, a, b, c=0, d=Dummy):
        self.b = b
        self.d = d
        self._input_args = tuple([a, weakref.ref(b)])
        self._input_kwargs = {'c': c, 'd': weakref.ref(d)}
        
    def __eq__(self, solf):
        if not isinstance(solf, self.__class__):  # pragma: debug
            return False
        return ((self._input_args == solf._input_args)
                and (self._input_kwargs == solf._input_kwargs))


class ValidArgsClass2(ValidArgsClass1):
    def get_input_args(self):
        return self._input_args

    def get_input_kwargs(self):
        return self._input_kwargs
    

class ValidArgsClass3(ValidArgsClass2):
    @property
    def input_arguments(self):
        return self._input_args

    @property
    def input_keyword_arguments(self):
        return self._input_kwargs


class ValidArgsClass4(ValidArgsClass1):

    def __init__(self, *args, **kwargs):
        self._input_args = args
        self._input_kwargs = kwargs


class InvalidArgsClass:  # pragma: no cover
    # Old style class dosn't inherit from object
    pass


class TestArgsMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for ArgsMetaschemaProperty class."""
    
    _mod = 'ArgsMetaschemaProperty'
    _cls = 'ArgsMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestArgsMetaschemaProperty, self).__init__(*args, **kwargs)
        self._valid = []
        self._invalid = []
        for cls in [ValidArgsClass1, ValidArgsClass2,
                    ValidArgsClass3, ValidArgsClass4]:
            cls_inst = cls(*(cls.test_args), **(cls.test_kwargs))
            self._valid.append((cls_inst, cls.valid_args))
            self._invalid.append((cls_inst, cls.invalid_args))
        valid_type = ValidArgsClass1.valid_args
        invalid_type = ValidArgsClass1.invalid_args
        self._encode_errors = [int(1), InvalidArgsClass]
        self._valid_compare = [(valid_type, valid_type)]
        self._invalid_compare = [(valid_type, invalid_type)]



import os
import sys
import argparse
import uuid
import pprint
import shutil
import subprocess
PYVER = ('%s.%s' % sys.version_info[:2])
PY2 = (sys.version_info[0] == 2)
_is_osx = (sys.platform == 'darwin')
_is_linux = ('linux' in sys.platform)
_is_win = (sys.platform in ['win32', 'cygwin'])
INSTALLLPY = (os.environ.get('INSTALLLPY', '0') == '1')
if _is_win:
    INSTALLC = (os.environ.get('INSTALLC', '0') == '1')
else:
    INSTALLC = True  # c compiler usually installed by default
INSTALLR = (os.environ.get('INSTALLR', '0') == '1')
INSTALLFORTRAN = (os.environ.get('INSTALLFORTRAN', '0') == '1')
if not INSTALLC:
    INSTALLFORTRAN = False
INSTALLSBML = (os.environ.get('INSTALLSBML', '0') == '1')
INSTALLAPY = (os.environ.get('INSTALLAPY', '0') == '1')
INSTALLZMQ = (os.environ.get('INSTALLZMQ', '0') == '1')
INSTALLRMQ = (os.environ.get('INSTALLRMQ', '0') == '1')
INSTALLTRIMESH = (os.environ.get('INSTALLTRIMESH', '0') == '1')
INSTALLPYGMENTS = (os.environ.get('INSTALLPYGMENTS', '0') == '1')
BUILDDOCS = (os.environ.get('BUILDDOCS', '0') == '1')


def call_conda_command(args, **kwargs):
    r"""Function for calling conda commands as the conda script is not
    available on subprocesses for windows unless invoked via the shell.

    Args:
        args (list): Command arguments.
        **kwargs: Additional keyword arguments are passed to subprocess.check_output.

    Returns:
        str: The output from the command.

    """
    if _is_win:
        args = ' '.join(args)
        kwargs['shell'] = True
    return subprocess.check_output(args, **kwargs).decode("utf-8")


def call_script(lines):
    r"""Write lines to a script and call it.

    Args:
        lines (list): Lines that should be written to the script.

    """
    if not lines:
        return
    if _is_win:  # pragma: windows
        script_ext = '.bat'
        error_check = 'if %errorlevel% neq 0 exit /b %errorlevel%'
        for i in range(len(lines), 0, -1):
            lines.insert(i, error_check)
    else:
        script_ext = '.sh'
        if lines[0] != '#!/bin/bash':
            lines.insert(0, '#!/bin/bash')
        error_check = 'set -e'
        if error_check not in lines:
            lines.insert(1, error_check)
    fname = 'ci_script_%s%s' % (str(uuid.uuid4()), script_ext)
    try:
        pprint.pprint(lines)
        with open(fname, 'w') as fd:
            fd.write('\n'.join(lines))
            
        call_kws = {}
        if _is_win:  # pragma: windows
            call_cmd = [fname]
        else:
            call_cmd = ['./%s' % fname]
            os.chmod(fname, 0o755)
        subprocess.check_call(call_cmd, **call_kws)
    finally:
        if os.path.isfile(fname):
            os.remove(fname)


def setup_package_on_ci(method, python):
    r"""Setup a test environment on a CI resource.

    Args:
        method (str): Method that should be used to build and install
            the package. Valid values include 'conda' and 'pip'.
        python (str): Version of Python that should be tested.

    Raises:
        ValueError: If method is not 'conda' or 'pip'.

    """
    cmds = []
    major, minor = [int(x) for x in python.split('.')]
    if os.environ.get('GITHUB_ACTIONS', False):
        conda_cmd = '$CONDA/bin/conda'
    elif _is_win:
        conda_cmd = 'call conda'
    else:
        conda_cmd = 'conda'
    if method == 'conda':
        cmds += [
            "echo Installing Python using conda...",
            # Configure conda
            "%s config --set always_yes yes --set changeps1 no" % conda_cmd,
            # "%s config --set channel_priority strict" % conda_cmd,
            "%s config --add channels conda-forge" % conda_cmd,
            "%s update -q conda" % conda_cmd,
            # "%s config --set allow_conda_downgrades true" % conda_cmd,
            # "%s install -n root conda=4.9" % conda_cmd,
            "%s create -q -n test-environment python=%s" % (conda_cmd, python)
        ]
    elif method == 'pip':
        if INSTALLLPY or _is_win:
            setup_package_on_ci('conda', python)
        elif _is_osx:
            cmds.append("echo Installing Python using virtualenv...")
            pip_cmd = 'pip'
            if sys.version_info[0] != major:
                pip_cmd = 'pip%d' % major
                try:
                    call_script(['python%d --version' % major])
                except BaseException:
                    cmds.append('brew install python%d' % major)
            cmds += [
                "%s install --upgrade pip virtualenv" % pip_cmd,
                "virtualenv -p python%d venv" % major
            ]
    else:  # pragma: debug
        raise ValueError("Method must be 'conda' or 'pip', not '%s'"
                         % method)
    call_script(cmds)
    

def deploy_package_on_ci(method, verbose=False):
    r"""Build and install the package and its dependencies on a CI
    resource.

    Args:
        method (str): Method that should be used to build and install
            the package. Valid values include 'conda' and 'pip'.
        verbose (bool, optional): If True, setup steps are run with verbosity
            turned up. Defaults to False.

    Raises:
        ValueError: If method is not 'conda' or 'pip'.

    """
    if _is_win:
        conda_cmd = 'call conda'
    else:
        conda_cmd = 'conda'
    cmds = [
        # Check that we have the expected version of Python
        "python --version",
    ]
    # Upgrade pip and setuptools and wheel to get clean install
    upgrade_pkgs = ['wheel', 'setuptools']
    if not _is_win:
        upgrade_pkgs.insert(0, 'pip')
    cmds += ["pip install --upgrade %s" % ' '.join(upgrade_pkgs)]
    # Uninstall default numpy and matplotlib to allow installation
    # of specific versions
    cmds += ["pip uninstall -y numpy matplotlib"]
    # Get dependencies
    install_req = os.path.join("utils", "install_from_requirements.py")
    conda_pkgs = []
    pip_pkgs = []
    requirements_files = ['requirements_testing.txt']
    _in_conda = False
    if method == 'conda':
        _in_conda = True
        default_pkgs = conda_pkgs
    elif method == 'pip':
        _in_conda = (_is_win or INSTALLLPY)
        default_pkgs = pip_pkgs
    else:  # pragma: debug
        raise ValueError("Method must be 'conda' or 'pip', not '%s'"
                         % method)
    # Installing via pip causes import error on Windows and
    # a conflict when installing LPy
    conda_pkgs += ['scipy', os.environ.get('NUMPY', 'numpy')]
    default_pkgs += [os.environ.get('MATPLOTLIB', 'matplotlib'),
                     os.environ.get('JSONSCHEMA', 'jsonschema')]
    if INSTALLSBML:
        pip_pkgs.append('libroadrunner')
    if INSTALLAPY:
        default_pkgs.append('astropy')
    if INSTALLRMQ:
        default_pkgs.append("\"pika<1.0.0b1\"")
    if INSTALLTRIMESH:
        default_pkgs.append('trimesh')
    if INSTALLPYGMENTS:
        default_pkgs.append('pygments')
    if (method == 'pip') and _in_conda:
        if INSTALLR:
            conda_pkgs.append('r-base')
        if INSTALLFORTRAN:
            if _is_win:
                conda_pkgs += ['m2w64-gcc-fortran', 'make',
                               'm2w64-toolchain_win-64']
            else:
                conda_pkgs += ['fortran-compiler']
        if INSTALLZMQ:
            conda_pkgs += ['czmq', 'zeromq']
    if BUILDDOCS:
        requirements_files.append('requirements_documentation.txt')
    # Install dependencies
    if method == 'conda':
        cmds += [
            "%s install -q -n base conda-build conda-verify" % conda_cmd,
        ]
    elif method == 'pip':
        if INSTALLR and (not _in_conda):
            cmds.append("echo Installing R...")
            if _is_linux:
                cmds += [("sudo add-apt-repository 'deb https://cloud"
                          ".r-project.org/bin/linux/ubuntu xenial-cran35/'"),
                         ("sudo apt-key adv --keyserver keyserver.ubuntu.com "
                          "--recv-keys E298A3A825C0D65DFD57CBB651716619E084DAB9"),
                         "sudo apt update",
                         "sudo apt-get install r-base r-base-dev",
                         "sudo apt-get install libudunits2-dev"]
            elif _is_osx:
                cmds += ["brew install r",
                         "brew install udunits"]
            else:
                raise NotImplementedError("Could not determine "
                                          "R installation method.")
        if INSTALLFORTRAN and (not _in_conda):
            cmds.append("echo Installing Fortran...")
            if _is_linux:
                cmds += ["sudo apt-get install gfortran"]
            elif _is_osx:
                cmds += ["brew install gfortran"]
            else:
                raise NotImplementedError("Could not determine "
                                          "Fortran installation method.")
        if INSTALLZMQ and (not _in_conda):
            cmds.append("echo Installing ZeroMQ...")
            if _is_linux:
                cmds.append("./ci/install-czmq-linux.sh")
            elif _is_osx:
                cmds.append("bash ci/install-czmq-osx.sh")
            # elif _is_win:
            #     cmds += ["call ci\\install-czmq-windows.bat",
            #              "echo \"%PATH%\""]
            else:
                raise NotImplementedError("Could not determine "
                                          "ZeroMQ installation method.")
    if _in_conda:
        if conda_pkgs:
            cmds += ["%s install %s" % (conda_cmd, ' '.join(conda_pkgs))]
    else:
        pip_pkgs += conda_pkgs
    if pip_pkgs:
        cmds += ["pip install %s" % ' '.join(pip_pkgs)]
    for x in requirements_files:
        cmds += ["python %s %s %s" % (install_req, method, x)]
    if INSTALLLPY:
        if _in_conda:
            cmds += ["%s install openalea.lpy boost=1.66.0 -c openalea"
                     % conda_cmd]
        else:  # pragma: debug
            raise RuntimeError("Could not detect conda environment. "
                               "Cannot proceed with a conda deployment "
                               "(required for LPy).")
    # Install yggdrasil
    if method == 'conda':
        # Install from conda build
        # Assumes that an environment is active
        conda_prefix = os.environ.get('CONDA_PREFIX', None)
        if not conda_prefix:
            if os.environ.get('GITHUB_ACTIONS', False):
                conda_prefix = os.environ['CONDA']
            else:
                conda_prefix = shutil.which('conda')
        prefix_dir = os.path.dirname(os.path.dirname(conda_prefix))
        index_dir = os.path.join(prefix_dir, "conda-bld")
        if verbose:
            build_flags = ''
            install_flags = '-vvv'
        else:
            build_flags = '-q'
            install_flags = '-q'
        cmds += [
            "%s build %s --python %s %s" % (
                conda_cmd, 'recipe', PYVER, build_flags),
            "%s index %s" % (conda_cmd, index_dir),
            "%s install %s --use-local --update-deps yggdrasil" % (
                conda_cmd, install_flags),
            # "%s install %s --update-deps -c file:/%s/conda-bld yggdrasil" % (
            #     conda_cmd, install_flags, prefix_dir),
        ]
    elif method == 'pip':
        if verbose:
            build_flags = ''
            install_flags = '--verbose'
        else:
            build_flags = '--quiet'
            install_flags = ''
        # Install from source dist
        cmds += ["python setup.py %s sdist" % build_flags]
        if _is_win:  # pragma: windows
            cmds += [
                "for %%a in (\"dist\\*.tar.gz\") do set YGGSDIST=%%a",
                "echo %YGGSDIST%"
            ]
            sdist = "%YGGSDIST%"
        else:
            sdist = "dist/*.tar.gz"
        cmds += [
            "pip install %s %s" % (install_flags, sdist),
            "python create_coveragerc.py"
        ]
    else:  # pragma: debug
        raise ValueError("Method must be 'conda' or 'pip', not '%s'"
                         % method)
    # Print summary of what was installed
    cmds.append('pip list')
    if _in_conda:
        cmds.append("%s list" % conda_cmd)
    call_script(cmds)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "Perform setup operations to test package build and "
        "installation on continuous integration services.")
    parser.add_argument(
        'method', choices=['conda', 'pip'],
        help=("Method that should be used to build "
              "and install the packages and its "
              "dependencies."))
    subparsers = parser.add_subparsers(
        dest='operation',
        help="CI setup operation that should be performed.")
    parser_env = subparsers.add_parser(
        'env', help="Setup an environment for testing.")
    parser_env.add_argument(
        'python',
        help="Version of python that should be tested.")
    parser_dep = subparsers.add_parser(
        'deploy', help="Build and install package.")
    parser_dep.add_argument(
        '--verbose', action='store_true',
        help="Turn up verbosity of output from setup step.")
    args = parser.parse_args()
    if args.operation in ['env', 'setup']:
        setup_package_on_ci(args.method, args.python)
    elif args.operation == 'deploy':
        deploy_package_on_ci(args.method, verbose=args.verbose)



from yggdrasil.communication.transforms.TransformBase import TransformBase


class DirectTransform(TransformBase):
    r"""Base class for message transforms."""

    _transformtype = 'direct'

    def evaluate_transform(self, x, no_copy=False):
        r"""Call transform on the provided message.

        Args:
            x (object): Message object to transform.
            no_copy (bool, optional): If True, the transformation occurs in
                place. Otherwise a copy is created and transformed. Defaults
                to False.

        Returns:
            object: The transformed message.

        """
        return x

    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the transform class.

        Returns:
            list: Multiple dictionaries of keywords and messages before/after
                pairs that will result from the transform created by the provided
                keywords.
        
        """
        return [{'kwargs': {}, 'in/out': [(1, 1)]}]



import os
import argparse
from yggdrasil import units
from yggdrasil.languages.Python.YggInterface import YggOutput, YggTimesync


def run(tmin, tmax, tstep):
    mass_out = YggOutput('mass')
    plant2root = YggTimesync('plant2root')
    tmin = units.add_units(tmin, 'days')
    tmax = units.add_units(tmax, 'days')
    tstep = units.add_units(tstep, 'days')
    t = tmin
    mass = 0

    while t <= tmax:

        # Calculate mass for the time step
        # (pretend this is a biologically complex calculation)
        mass += t * units.add_units(0.2, 'kg/days')

        # Synchronize data for time step with the root model
        plant_state = {'mass': mass}
        flag, plant_state = plant2root.call(t, plant_state)
        if not flag:
            raise Exception("Error performing time-step synchronization "
                            "with plant model.")
    
        # Send masses to output
        flag = mass_out.send({'time': t,
                              'total_mass': plant_state['mass'],
                              'root_mass': mass,
                              'plant_mass': plant_state['mass'] - mass})
        if not flag:
            raise Exception("Error sending masses to output")

        # Advance time step
        t += tstep

    return mass


# Parse command-line arguments
if __name__ == "__main__":
    parser = argparse.ArgumentParser("Simulate root growth over time.")
    parser.add_argument('tmin', help='Starting time (in days)', type=float)
    parser.add_argument('tmax', help='Ending time (in days)', type=float)
    parser.add_argument('tstep', help='Time step (in days)', type=float)
    args = parser.parse_args()
    run(args.tmin, args.tmax, args.tstep)



import numpy as np
import pandas
import copy
from yggdrasil import serialize
from yggdrasil.communication.transforms.TransformBase import TransformBase


class SelectFieldsTransform(TransformBase):
    r"""Class for selecting a subset of the original fields in an object.

    Args:
        selected (list): A list of fields that should be selected.
        original_order (list, optional): The original order of fields that
            should be used for selecting from lists/tuples.
        single_as_scalar (bool, optional): If True and only a single field
            is selected, the transformed messages will be scalars rather
            than arrays with single elements. Defaults to False.

    """
    _transformtype = 'select_fields'
    _schema_required = ['selected']
    _schema_properties = {'selected': {'type': 'array',
                                       'items': {'type': 'string'}},
                          'original_order': {'type': 'array',
                                             'items': {'type': 'string'}},
                          'single_as_scalar': {'type': 'boolean'}}

    def set_original_datatype(self, datatype):
        r"""Set datatype.

        Args:
            datatype (dict): Datatype.

        """
        super(SelectFieldsTransform, self).set_original_datatype(datatype)
        if not self.original_order:
            self.original_order = self.original_datatype.get('field_names', None)
        if not self.original_order:
            if (((datatype['type'] == 'array')
                 and isinstance(datatype['items'], list))):
                self.original_order = [x.get('title', 'f%d' % i) for i, x in
                                       enumerate(self.original_datatype['items'])]
            elif datatype['type'] == 'object':
                self.original_order = list(datatype['properties'].keys())

    @property
    def as_single(self):
        r"""bool: True if there is a single element to return."""
        return (self.single_as_scalar and (len(self.selected) == 1))

    def validate_datatype(self, datatype):
        r"""Assert that the provided datatype is valid for this transformation.
        
        Args:
            datatype (dict): Datatype to validate.

        Raises:
            AssertionError: If the datatype is not valid.

        """
        assert(datatype.get('type', None) in ['array', 'object'])
        
    def transform_datatype(self, datatype):
        r"""Determine the datatype that will result from applying the transform
        to the supplied datatype.

        Args:
            datatype (dict): Datatype to transform.

        Returns:
            dict: Transformed datatype.

        """
        if (((datatype.get('type', None) == 'array')
             and isinstance(datatype.get('items', None), list))):
            order = datatype.get('field_names',
                                 [x.get('title', 'f%d' % i)
                                  for i, x in enumerate(datatype['items'])])
            if self.as_single:
                datatype = copy.deepcopy(datatype['items'][
                    order.index(self.selected[0])])
                datatype['title'] = self.selected[0]
            else:
                datatype = copy.deepcopy(datatype)
                datatype['items'] = [datatype['items'][order.index(k)]
                                     for k in self.selected]
                for i, k in enumerate(self.selected):
                    datatype['items'][i]['title'] = k
                if 'field_names' in datatype:
                    datatype['field_names'] = copy.deepcopy(self.selected)
                if 'format_str' in datatype:
                    info = serialize.format2table(datatype['format_str'])
                    info['fmts'] = [info['fmts'][order.index(k)]
                                    for k in self.selected]
                    datatype['format_str'] = serialize.table2format(**info)
        elif (((datatype.get('type', None) == 'array')
               and isinstance(datatype.get('items', None), dict)
               and self.as_single)):
            datatype = copy.deepcopy(datatype['items'])
        elif datatype.get('type', None) == 'object':
            if self.as_single:
                datatype = copy.deepcopy(datatype['properties'][self.selected[0]])
            else:
                datatype = copy.deepcopy(datatype)
                datatype['properties'] = {k: datatype['properties'][k]
                                          for k in self.selected}
        return datatype
    
    def evaluate_transform(self, x, no_copy=False):
        r"""Call transform on the provided message.

        Args:
            x (object): Message object to transform.
            no_copy (bool, optional): If True, the transformation occurs in
                place. Otherwise a copy is created and transformed. Defaults
                to False.

        Returns:
            object: The transformed message.

        """
        out = x
        if isinstance(x, dict):
            if self.as_single:
                out = x[self.selected[0]]
            else:
                out = type(x)([(k, x[k]) for k in self.selected])
        elif isinstance(x, (list, tuple)):
            if not self.original_order:
                raise ValueError("The original order of the fields must be "
                                 "provided for list/tuple objects.")
            if self.as_single:
                out = x[self.original_order.index(self.selected[0])]
            else:
                out = type(x)([x[self.original_order.index(k)]
                               for k in self.selected])
        elif isinstance(x, (np.ndarray, pandas.DataFrame)):
            if self.as_single:
                out = x[self.selected[0]]
            else:
                out = x[self.selected]
        else:
            raise TypeError("Cannot select fields from object of type '%s'" % type(x))
        if not no_copy:
            out = copy.deepcopy(out)
        return out
    
    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the transform class.

        Returns:
            list: Multiple dictionaries of keywords and messages before/after
                pairs that will result from the transform created by the provided
                keywords.
        
        """
        return [{'kwargs': {'selected': ['a', 'c'],
                            'original_datatype': {
                                'type': 'object',
                                'properties': {x: {'type': 'int'}
                                               for x in 'abc'}}},
                 'in/out': [(dict(zip('abc', range(3))), {'a': 0, 'c': 2})],
                 'in/out_t': [({'type': 'object',
                                'properties': {x: {'type': 'int'}
                                               for x in 'abc'}},
                               {'type': 'object',
                                'properties': {x: {'type': 'int'}
                                               for x in 'ac'}})]},
                {'kwargs': {'selected': ['a'],
                            'single_as_scalar': True,
                            'original_datatype': {
                                'type': 'object',
                                'properties': {x: {'type': 'int'}
                                               for x in 'abc'}}},
                 'in/out': [(dict(zip('abc', range(3))), 0)],
                 'in/out_t': [({'type': 'object',
                                'properties': {x: {'type': 'int'}
                                               for x in 'abc'}},
                               {'type': 'int'})]},
                {'kwargs': {'selected': ['a', 'c'],
                            'original_datatype': {
                                'type': 'array',
                                'items': {'type': 'int'}}},
                 'in/out': [([0, 1, 2], ValueError)]},
                {'kwargs': {'selected': ['a', 'c'],
                            'original_datatype': {
                                'type': 'array',
                                'items': [
                                    {'type': 'int', 'title': x}
                                    for x in 'abc']}},
                 'in/out': [([0, 1, 2], [0, 2])],
                 'in/out_t': [({'type': 'array',
                                'items': [
                                    {'type': 'int', 'title': x}
                                    for x in 'abc'],
                                'format_str': b'# %d\t%d\t%d\n'},
                               {'type': 'array',
                                'items': [
                                    {'type': 'int', 'title': x}
                                    for x in 'ac'],
                                'format_str': b'# %d\t%d\n'}),
                              ({'type': 'array',
                                'field_names': [x for x in 'abc'],
                                'items': [
                                    {'type': 'int', 'title': x}
                                    for x in 'abc']},
                               {'type': 'array',
                                'field_names': [x for x in 'ac'],
                                'items': [
                                    {'type': 'int', 'title': x}
                                    for x in 'ac']})]},
                {'kwargs': {'selected': ['a', 'c'],
                            'original_order': ['a', 'b', 'c'],
                            'original_datatype': {
                                'type': 'array',
                                'items': {'type': 'int'}}},
                 'in/out': [([0, 1, 2], [0, 2])],
                 'in/out_t': [({'type': 'array',
                                'items': {'type': 'int'}},
                               {'type': 'array',
                                'items': {'type': 'int'}})]},
                {'kwargs': {'selected': ['a'],
                            'single_as_scalar': True,
                            'original_datatype': {
                                'type': 'array',
                                'items': [
                                    {'type': 'int', 'title': x}
                                    for x in 'abc']}},
                 'in/out': [([0, 1, 2], 0)],
                 'in/out_t': [({'type': 'array',
                                'items': [
                                    {'type': 'int', 'title': x}
                                    for x in 'abc']},
                               {'type': 'int', 'title': 'a'})]},
                {'kwargs': {'selected': ['a'],
                            'single_as_scalar': True,
                            'original_order': ['a', 'b', 'c'],
                            'original_datatype': {
                                'type': 'array',
                                'items': {'type': 'int'}}},
                 'in/out': [([0, 1, 2], 0)],
                 'in/out_t': [({'type': 'array',
                                'items': {'type': 'int'}},
                               {'type': 'int'})]},
                {'kwargs': {'selected': ['a', 'c']},
                 'in/out': [(np.zeros(3, np.dtype({'names': ['a', 'b', 'c'],
                                                   'formats': ['i4', 'i4', 'i4']})),
                             np.zeros(3, np.dtype({'names': ['a', 'c'],
                                                   'formats': ['i4', 'i4']})))]},
                {'kwargs': {'selected': ['a'],
                            'single_as_scalar': True},
                 'in/out': [(np.zeros(3, np.dtype({'names': ['a', 'b', 'c'],
                                                   'formats': ['i4', 'i4', 'i4']})),
                             np.zeros(3, np.dtype('i4')))]},
                {'kwargs': {'selected': ['a', 'b'],
                            'original_datatype': {
                                'type': 'array',
                                'items': {'type': 'int'}}},
                 'in/out': [(None, TypeError)]}]



import os
import uuid
from yggdrasil.tests import generate_component_tests, assert_equal, assert_raises
from yggdrasil.communication import open_file_comm
from yggdrasil.communication.tests.test_FileComm import TestFileComm


generate_component_tests('file', TestFileComm, globals(), __file__,
                         class_attr='comm')


def test_open_file_comm():
    r"""Test file comm context manager."""
    fname = os.path.join(os.path.dirname(__file__),
                         '%s.txt' % str(uuid.uuid4()))
    try:
        msg = b'Test message'
        with open_file_comm(fname, 'w') as comm:
            flag = comm.send(msg)
            assert(flag)
        with open_file_comm(fname, 'a') as comm:
            flag = comm.send(msg)
            assert(flag)
        with open_file_comm(fname, 'r') as comm:
            flag, res = comm.recv()
            assert(flag)
        assert_equal(res, 2 * msg)
    finally:
        if os.path.isfile(fname):
            os.remove(fname)
    with assert_raises(ValueError):
        with open_file_comm(fname, 'invalid') as comm:  # pragma: debug
            pass


__all__ = []



import os
import argparse
from install_from_requirements import locate_conda_exe
from ci_setup import call_conda_command
_utils_dir = os.path.dirname(os.path.abspath(__file__))
_pkg_dir = os.path.dirname(_utils_dir)


def conda_env_exists(name):
    r"""Determine if a conda environment already exists.

    Args:
        name (str): Name of the environment to check.

    Returns:
        bool: True the the environment exits, False otherwise.

    """
    args = ['conda', 'info', '--envs']
    out = call_conda_command(args)
    envs = []
    for x in out.splitlines():
        if x.startswith('#') or (not x):
            continue
        envs.append(x.split()[0])
    return (name in envs)


def create_env(name, python='3.6', packages=None):
    r"""Create a new conda environment with the specified parameters.

    Args:
        name (str): Name of the environment being created.
        python (str, optional): Version of python that should be used. Defaults
            to '3.6'.
        packages (list, optional): Packages that should be installed in the new
            environment. Defaults to None and is ignored.

    """
    if conda_env_exists(name):
        print("Conda env with name '%s' already exists." % name)
        return
    if packages is None:
        packages = []
    if 'ipython' not in packages:
        packages.append('ipython')
    args = ['conda', 'create', '-y', '-n', name, 'python=%s' % python] + packages
    print('running', args)
    print(call_conda_command(args))
    assert(conda_env_exists(name))


def create_devenv(env_type, python='3.6', name=None, **kwargs):
    r"""Create a new conda environment with the specified parameters with
    yggdrasil installed in development mode.

    Args:
        env_type (str): Type of environment that should be created, either
            'pip' or 'conda'.
        python (str, optional): Version of python that should be used. Defaults
            to '3.6'.
        name (str, optional): Name that should be given to the environment.
            Defaults to None and a name will be created from the provided
            values for env_type and python.
        **kwargs: Additional keyword arguments are passed to create_env.

    """
    assert(env_type in ['pip', 'conda'])
    if name is None:
        name = '%s%s' % (env_type, python.replace('.', ''))
    create_env(name, python=python, **kwargs)
    python_cmd = locate_conda_exe(name, 'python')
    req_suffixes = ['', '_testing', '_development']
    if env_type == 'conda':
        req_suffixes.append('_condaonly')
        # TODO: Find a way to install pip-only packages into conda
        # environment
    elif env_type == 'pip':
        req_suffixes.append('_piponly')
        print(call_conda_command(['conda', 'install', '-y', '--name', name,
                                  'czmq', 'zeromq']))
    req = [os.path.join(_pkg_dir, 'requirements%s.txt' % x)
           for x in req_suffixes]
    args = ([python_cmd, os.path.join(_utils_dir,
                                      'install_from_requirements.py'),
             '--conda-env', name, env_type] + req)
    print(call_conda_command(args))
    # This is not called directly because it needs to be invoked with the
    # version of Python installed in the target environment
    # install_from_requirements(env_type, req, conda_env=name)
    print(call_conda_command([python_cmd, 'setup.py', 'develop'],
                             cwd=_pkg_dir))
    print("Created environment %s" % name)
    

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        "Create dev environments for matrix of installation methods "
        "and Python versions.")
    parser.add_argument('--method', choices=['conda', 'pip', 'both'], default='both',
                        help=("Method(s) that should be used to install the "
                              "dependencies."))
    parser.add_argument('--version', '--python', default='2&3',
                        choices=['2.7', '3.6', '3.7', '3.8', '2&3'],
                        help=("Python version(s) to create environments for."))
    parser.add_argument('--name', default=None,
                        help=("Name of the conda env that should be created or "
                              "updated. If provided, 'method' will default to "
                              "'conda' and 'version' will default to '3.6'."))
    args = parser.parse_args()
    if args.method == 'both':
        if args.name:
            args.method = ['conda']
        else:
            args.method = ['pip', 'conda']
    else:
        args.method = [args.method]
    if args.version == '2&3':
        if args.name:
            args.version = ['3.6']
        else:
            args.version = ['2.7', '3.6']
    else:
        args.version = [args.version]
    for env_type in args.method:
        for python in args.version:
            create_devenv(env_type, python=python, name=args.name)



import os
import re
import copy
import logging
from collections import OrderedDict
from yggdrasil import platform, tools
from yggdrasil.languages import get_language_dir
from yggdrasil.drivers import CModelDriver
from yggdrasil.drivers.CompiledModelDriver import (
    CompilerBase, CompiledModelDriver, get_compilation_tool)
from yggdrasil.metaschema.properties.ScalarMetaschemaProperties import (
    _valid_types)


logger = logging.getLogger(__name__)
_top_lang_dir = get_language_dir('fortran')
_incl_interface = _top_lang_dir
_c_internal_libs = copy.deepcopy(CModelDriver.CModelDriver.internal_libraries)


# TODO: Add support for f77: e.g.
# https://people.sc.fsu.edu/~jburkardt/f77_src/f77_calls_c/f77_calls_c.html
class FortranCompilerBase(CompilerBase):
    r"""Base class for Fortran compilers."""
    languages = ['fortran']
    default_executable_env = 'FC'
    default_flags_env = 'FFLAGS'
    default_flags = ['-g', '-Wall', '-cpp', '-pedantic-errors', '-ffree-line-length-0']
    linker_attributes = {'default_flags_env': 'LFLAGS',
                         'search_path_envvar': ['LIBRARY_PATH', 'LD_LIBRARY_PATH']}
    search_path_envvar = []
    default_linker = None
    default_executable = None
    default_archiver = None
    product_exts = ['mod']

    @classmethod
    def get_flags(cls, **kwargs):
        r"""Get a list of flags for the tool.

        Args:
            **kwargs: Keyword arguments are passed to the parent class's
                method.

        Returns:
            list: Flags for the tool.

        """
        if 'module-dir' in cls.flag_options:
            kwargs.setdefault('module-dir', _top_lang_dir)
        if 'module-search-path' in cls.flag_options:
            kwargs.setdefault('module-search-path', _top_lang_dir)
        kwargs.setdefault('include_dirs', cls.get_search_path())
        out = super(FortranCompilerBase, cls).get_flags(**kwargs)
        for x in ['-O', '-O2', '-O3', 'Os', 'Ofast']:  # pragma: debug
            if x in out:
                out.remove(x)
        return out
        
    @classmethod
    def append_product(cls, products, src, new, new_dir=None,
                       dont_append_src=False):
        r"""Append a product to the specified list along with additional values
        indicated by cls.product_exts.

        Args:
            products (list): List of of existing products that new product
                should be appended to.
            src (list): Input arguments to compilation call that was used to
                generate the output file (usually one or more source files).
            new (str): New product that should be appended to the list.
            new_dir (str, optional): Directory that should be used as base when
                adding files listed in cls.product_files. Defaults to
                os.path.dirname(new).
            dont_append_src (bool, optional): If True and src is in the list of
                products, it will be removed. Defaults to False.

        """
        super(FortranCompilerBase, cls).append_product(
            products, src, new, new_dir=new_dir,
            dont_append_src=dont_append_src)
        if os.path.basename(new).startswith('YggInterface_f90'):
            products.append(os.path.join(_top_lang_dir,
                                         'fygg.mod'))


# class FlangCompiler(FortranCompilerBase):
#     r"""Interface class for flang compiler/linker."""
#     toolname = 'flang'
#     platforms = ['MacOS', 'Linux', 'Windows']
#     default_flags = (FortranCompilerBase.default_flags
#                      + ['-Werror', '-w', '-Weverything'])
#     flag_options = OrderedDict(list(FortranCompilerBase.flag_options.items())
#                                + [('module-dir', '-I%s'),
#                                   ('module-search-path', '-I%s'),
#                                   ('standard', '-std=%s')])
#     toolset = 'llvm'


class GFortranCompiler(FortranCompilerBase):
    r"""Interface class for gfortran compiler/linker."""
    toolname = 'gfortran'
    platforms = ['MacOS', 'Linux', 'Windows']
    default_flags = (FortranCompilerBase.default_flags
                     + ['-x', 'f95-cpp-input'])
    flag_options = OrderedDict(list(FortranCompilerBase.flag_options.items())
                               + [('module-dir', '-J%s'),
                                  ('module-search-path', '-I%s'),
                                  ('standard', '-std=%s')])
    toolset = 'gnu'
    compatible_toolsets = ['llvm']
    default_archiver = 'ar'


# class IFortCompiler(FortranCompilerBase):
#     r"""Interface class for ifort compiler/linker."""
#     toolname = 'ifort'
#     platforms = ['MacOS', 'Linux', 'Windows']
#     flag_options = OrderedDict(list(FortranCompilerBase.flag_options.items())
#                                + [('module-dir', '-module'),
#                                   ('module-search-path', '-module'),
#                                   ('standard', '-stand')])
#     toolset = 'msvc'  # Is this strictly true?


class FortranModelDriver(CompiledModelDriver):
    r"""Class for running Fortran models.

    Args:
        standard (str, optional): Fortran standard that should be used.
            Defaults to 'f2003'.
        **kwargs: Additional keyword arguments are passed to parent class.

    """
                
    _schema_subtype_description = ('Model is written in Fortran.')
    _schema_properties = {'standard': {'type': 'string',
                                       'default': 'f2003',
                                       'enum': ['f2003', 'f2008']}}
    language = 'fortran'
    language_ext = ['.f90', '.f77', '.f', '.h']
    base_languages = ['c']
    interface_library = 'fygg'
    # To prevent inheritance
    default_compiler = 'gfortran'
    default_linker = None
    supported_comm_options = dict(
        CModelDriver.CModelDriver.supported_comm_options,
        zmq={'libraries': [('c', x) for x in
                           CModelDriver.CModelDriver.supported_comm_options[
                               'zmq']['libraries']]})
    external_libraries = {'cxx': {'include': 'stdlib.h',
                                  'libtype': 'shared',
                                  'language': 'c'}}
    internal_libraries = dict(
        fygg={'source': os.path.join(_incl_interface,
                                     'YggInterface.f90'),
              'libtype': 'static',
              'internal_dependencies': (
                  [('c', x) for x in
                   _c_internal_libs['ygg']['internal_dependencies']]
                  + [('c', 'ygg'), 'c_wrappers']),
              'external_dependencies': (
                  [('c', x) for x in
                   _c_internal_libs['ygg']['external_dependencies']]),
              'include_dirs': (
                  _c_internal_libs['ygg']['include_dirs'])},
        c_wrappers={'source': os.path.join(_incl_interface,
                                           'c_wrappers.c'),
                    'language': 'c',
                    'libtype': 'object',
                    'internal_dependencies': [('c', 'ygg')],
                    'external_dependencies': (
                        [('c', x) for x in
                         _c_internal_libs['ygg']['external_dependencies']]),
                    'include_dirs': [_incl_interface]})
    type_map = {
        'comm': 'yggcomm',
        'dtype': 'yggdtype',
        'int': 'integer(kind = X)',
        'float': 'real(kind = X)',
        'string': 'character(len = X)',
        'array': 'yggarr',
        'object': 'yggmap',
        'integer': 'integer',
        'boolean': 'logical(kind = X)',
        'null': 'yggnull',
        'uint': 'ygguintX',  # Fortran has no unsigned int
        'complex': 'complex(kind = X)',
        'bytes': 'character(len = X)',
        'unicode': ('character(kind = selected_char_kind(\'ISO_10646\'), '
                    'len = X)'),
        '1darray': '*',
        'ndarray': '*',
        '1darray_pointer': '{type}{precision}_1d',
        'ndarray_pointer': '{type}{precision}_{ndim}d',
        'ply': 'yggply',
        'obj': 'yggobj',
        'schema': 'yggschema',
        'flag': 'logical',
        'class': 'yggpython',
        'function': 'yggpyfunc',
        'instance': 'yggpyinst',
        'any': 'ygggeneric'}
    function_param = {
        'import_nofile': 'use {function}',
        'import': '#include "{filename}"',
        'index': '{variable}({index})',
        'interface': 'use fygg',
        'input': ('{channel} = ygg_input_type('
                  '\"{channel_name}\", {channel_type})'),
        'output': ('{channel} = ygg_output_type('
                   '\"{channel_name}\", {channel_type})'),
        'recv_heap': 'ygg_recv_var_realloc',
        'recv_stack': 'ygg_recv_var',
        'recv_function': 'ygg_recv_var_realloc',
        'send_function': 'ygg_send_var',
        'not_flag_cond': '.not.{flag_var}',
        'flag_cond': '{flag_var}',
        'declare': '{type_name} :: {variable}',
        'init_array': 'yggarr(init_generic())',
        'init_object': 'yggmap(init_generic())',
        'init_schema': 'yggschema(init_generic())',
        'init_ply': 'init_ply()',
        'init_obj': 'init_obj()',
        'init_class': 'init_python()',
        'init_function': 'yggpyfunc(init_python())',
        'init_instance': 'yggpyinst(init_generic())',
        'init_any': 'init_generic()',
        'init_type_array': ('create_dtype_json_array({nitems}, '
                            '{items}, {use_generic})'),
        'init_type_object': ('create_dtype_json_object({nitems}, '
                             '{keys}, {values}, {use_generic})'),
        'init_type_ply': 'create_dtype_ply({use_generic})',
        'init_type_obj': 'create_dtype_obj({use_generic})',
        'init_type_1darray': ('create_dtype_1darray(\"{subtype}\", '
                              '{precision}, {length}, \"{units}\", '
                              '{use_generic})'),
        'init_type_ndarray': ('create_dtype_ndarray(\"{subtype}\", '
                              '{precision}, {ndim}, {shape}, '
                              '\"{units}\", {use_generic})'),
        'init_type_ndarray_arr': ('create_dtype_ndarray(\"{subtype}\", '
                                  '{precision}, {ndim}, {shape}, '
                                  '\"{units}\", {use_generic})'),
        'init_type_scalar': ('create_dtype_scalar(\"{subtype}\", '
                             '{precision}, \"{units}\", '
                             '{use_generic})'),
        'init_type_default': ('create_dtype_default(\"{type}\", '
                              '{use_generic})'),
        'init_type_pyobj': ('create_dtype_pyobj(\"{type}\", '
                            '{use_generic})'),
        'init_type_empty': ('create_dtype_empty({use_generic})'),
        'init_type_schema': ('create_dtype_schema({use_generic})'),
        'copy_array': '{name} = yggarr(copy_generic(ygggeneric({value})))',
        'copy_object': '{name} = yggmap(copy_generic(ygggeneric({value})))',
        'copy_schema': '{name} = yggschema(copy_generic(ygggeneric({value})))',
        'copy_ply': '{name} = copy_ply({value})',
        'copy_obj': '{name} = copy_obj({value})',
        'copy_class': '{name} = copy_python({value})',
        'copy_function': '{name} = yggpyfunc(copy_python(yggpython({value})))',
        'copy_instance': '{name} = yggpyinst(copy_generic(ygggeneric({value})))',
        'copy_generic': '{name} = copy_generic({value})',
        'copy_any': '{name} = copy_generic({value})',
        'free_array': 'call free_generic(ygggeneric({variable}))',
        'free_object': 'call free_generic(ygggeneric({variable}))',
        'free_schema': 'call free_generic(ygggeneric({variable}))',
        'free_ply': 'call free_ply({variable})',
        'free_obj': 'call free_obj({variable})',
        'free_class': 'call free_python({variable})',
        'free_function': 'call free_python(yggpython({variable}))',
        'free_instance': 'call free_generic(ygggeneric({variable}))',
        'free_any': 'call free_generic({variable})',
        'print_generic': 'write(*, *) {object}',
        'print': 'write(*, \'(\"{message}\")\')',
        'fprintf': 'write(*, \'(\"{message}\")\') {variables}',
        'print_array': 'call display_generic(ygggeneric({object}))',
        'print_object': 'call display_generic(ygggeneric({object}))',
        'print_schema': 'call display_generic(ygggeneric({object}))',
        'print_ply': 'call display_ply({object})',
        'print_obj': 'call display_obj({object})',
        'print_class': 'call display_python({object})',
        'print_function': 'call display_python(yggpython({object}))',
        'print_instance': 'call display_generic(ygggeneric({object}))',
        'print_any': 'call display_generic({object})',
        'print_null': 'call display_null({object})',
        'assign': '{name} = {value}',
        'comment': '!',
        'true': '.true.',
        'false': '.false.',
        'null': 'c_null_ptr',
        'not': '.not.',
        'and': '.and.',
        'or': '.or.',
        'indent': 3 * ' ',
        'quote': "'",
        'error': ("write(*, \'(\"{error_msg}\")\')\n"
                  "stop 1"),
        'continuation_before': '&',
        'continuation_after': '     &',
        'block_end': 'END',
        'if_begin': 'IF ({cond}) THEN',
        'if_elif': 'ELSE IF ({cond}) THEN',
        'if_else': 'ELSE',
        'if_end': 'END IF',
        'for_begin': 'DO {iter_var} = {iter_begin}, {iter_end}',
        'for_end': 'END DO',
        'while_begin': 'DO WHILE ({cond})',
        'while_end': 'END DO',
        'break': 'EXIT',
        'exec_begin': 'PROGRAM main\n   use iso_c_binding\n   use fygg',
        'exec_end': '   stop\nEND PROGRAM main',
        'free': 'DEALLOCATE({variable})',
        'function_def_begin': (
            'FUNCTION {function_name}({input_var}) '
            'result({output_var})'),
        'function_def_end': 'END FUNCTION {function_name}',
        'subroutine_def_begin': (
            'SUBROUTINE {function_name}({input_var}) '),
        'subroutine_def_end': 'END SUBROUTINE {function_name}',
        'function_call_noout': 'call {function_name}({input_var})',
        'function_def_regex': (
            r'(?P<procedure_type>(?i:(?:subroutine)|(?:function)))\s+'
            r'{function_name}\s*\((?P<inputs>(?:[^\(]*?))\)\s*'
            r'(?:result\s*\((?P<flag_var>.+)\))?\s*\n'
            r'(?P<preamble>(?:[^:]*?\n)*?)'
            r'(?P<definitions>(?:(?:(?: )|(?:.))+\s*::\s*(?:.+)\n)+)'
            r'(?P<body>(?:.*?\n?)*?)'
            r'(?i:end\s+(?P=procedure_type))\s+{function_name}'),
        'definition_regex': (
            r'\s*(?P<type>(?:(?: )|(?:.))+)\s*::\s*(?P<name>.+)'
            r'(?:\s*=\s*(?P<value>.+))?\n'),
        'inputs_def_regex': (
            r'\s*(?P<name>.+?)\s*(?:,|$)(?:\n)?'),
        'outputs_def_regex': (
            r'\s*(?P<name>.+?)\s*(?:,|$)(?:\n)?'),
        'type_regex': (
            r'(type\()?(?P<type>[^,\(]+)(?(1)(?:\)))'
            r'(?:\s*\(\s*'
            r'(?:kind\s*=\s*(?:(?P<precision>\d*)|'
            r'(?P<precision_var>.+?)))?\s*,?\s*'
            r'(?:len\s*=\s*(?:(?P<length>(?:\d+))|'
            r'(?P<length_var>.+?)))?'
            r'\))?'
            r'(?:\s*,\s*dimension\((?:(?P<shape>(?:\d)+'
            r'(?:,\s*(?:\d)+)*?)'
            r'|(?P<shape_var>.+?(?:,\s*.+)*?))\))?'
            r'(?:\s*,\s*(?P<pointer>pointer))?'
            r'(?:\s*,\s*(?P<target>target))?'
            r'(?:\s*,\s*(?P<allocatable>allocatable))?'
            r'(?:\s*,\s*(?P<parameter>parameter))?'
            r'(?:\s*,\s*intent\((?P<intent>.*?)\))?')
    }
    outputs_in_inputs = True
    include_channel_obj = True
    is_typed = True
    types_in_funcdef = False
    interface_inside_exec = True
    zero_based = False
    max_line_width = 72
    global_scope_macro = ('#define WITH_GLOBAL_SCOPE(COMM) call '
                          'set_global_comm(); COMM; call unset_global_comm()')
    
    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        # if cls.default_compiler is None:
        #     if platform._is_linux or platform._is_mac:
        #         cls.default_compiler = 'gfortran'
        #     elif platform._is_win:  # pragma: windows
        #         cls.default_compiler = 'flang'
        CompiledModelDriver.before_registration(cls)
        cxx_orig = cls.external_libraries.pop('cxx', None)
        if cxx_orig is not None:
            if platform._is_mac:
                cxx_lib = 'c++'
            else:
                cxx_lib = 'stdc++'
            if cxx_lib not in cls.external_libraries:
                cls.external_libraries[cxx_lib] = cxx_orig.copy()
                cls.internal_libraries['fygg']['external_dependencies'].append(cxx_lib)
        if platform._is_win:  # pragma: windows
            cl_compiler = get_compilation_tool('compiler', 'cl')
            if not cl_compiler.is_installed():  # pragma: debug
                logger.info("The MSVC compiler could not be located. The Python C API "
                            "assumes the MSVC CRT will be used on windows so there may "
                            "be errors when accessing some behavior of the Python C API."
                            "In particular, segfaults are known to occur when trying to "
                            "call display_python due to differences in the internal "
                            "structure of FILE* objects between MSVC and other "
                            "standards.")

    @classmethod
    def set_env_class(cls, **kwargs):
        r"""Set environment variables that are instance independent.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method and update_ld_library_path.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(FortranModelDriver, cls).set_env_class(**kwargs)
        out = CModelDriver.CModelDriver.set_env_class(
            existing=out, add_libpython_dir=True, toolname=kwargs.get('toolname', None))
        out = CModelDriver.CCompilerBase.set_env(out)
        conda_prefix = tools.get_conda_prefix()
        if conda_prefix and platform._is_mac:
            out.setdefault('DYLD_FALLBACK_LIBRARY_PATH',
                           os.path.join(conda_prefix, 'lib'))
        return out

    def compile_model(self, **kwargs):
        r"""Compile model executable(s).

        Args:
            **kwargs: Keyword arguments are passed to the parent class's
            method.

        Returns:
            str: Compiled model file path.

        """
        kwargs.setdefault('standard', self.standard)
        return super(FortranModelDriver, self).compile_model(**kwargs)

    @classmethod
    def get_internal_suffix(cls, commtype=None):
        r"""Determine the suffix that should be used for internal libraries.

        Args:
            commtype (str, optional): If provided, this is the communication
                type that should be used for the model. If None, the
                default comm is used.

        Returns:
            str: Suffix that should be added to internal libraries to
                differentiate between different dependencies.

        """
        out = super(FortranModelDriver, cls).get_internal_suffix(
            commtype=commtype)
        if commtype is None:
            commtype = tools.get_default_comm()
        out += '_%s' % commtype[:3].lower()
        return out

    @classmethod
    def get_inverse_type_map(cls):
        r"""Get the inverse type map.

        Returns:
            dict: Mapping from native type to JSON type.

        """
        out = super(FortranModelDriver, cls).get_inverse_type_map()
        out['yggchar_r'] = 'bytes'
        return out
        
    @classmethod
    def get_native_type(cls, **kwargs):
        r"""Get the native type.

        Args:
            type (str, optional): Name of |yggdrasil| extended JSON
                type or JSONSchema dictionary defining a datatype.
            **kwargs: Additional keyword arguments may be used in determining
                the precise declaration that should be used.

        Returns:
            str: The native type.

        """
        out = super(FortranModelDriver, cls).get_native_type(**kwargs)
        intent_regex = r'(,\s*intent\(.+?\))'
        for x in re.finditer(intent_regex, out):
            out = out.replace(x.group(0), '')
        type_match = re.search(cls.function_param['type_regex'], out)
        if type_match:
            type_match = type_match.groupdict()
            if type_match.get('shape_var', None):  # pragma: debug
                if ('pointer' not in out) and ('allocatable' not in out):
                    out += ', allocatable'
                if type_match['shape_var'][0] == '*':
                    out = out.replace('*', ':')
                # raise Exception("Used default native_type, but need alias")
            elif type_match.get('length_var', None):
                if ((('pointer' not in out) and ('allocatable' not in out)
                     and (type_match['length_var'] != 'X'))):
                    out += ', allocatable'
                if type_match['length_var'] == '*':
                    out = out.replace('*', ':')
        if not ((out == '*') or ('X' in out)):
            if out.startswith('ygg'):
                out = 'type(%s)' % out
            return out
        from yggdrasil.metaschema.datatypes import get_type_class
        json_type = kwargs.get('datatype', kwargs.get('type', 'bytes'))
        if isinstance(json_type, str):  # pragma: no cover
            json_type = {'type': json_type}
        if 'type' in kwargs:  # pragma: no cover
            json_type.update(kwargs)
        assert(isinstance(json_type, dict))
        json_type = get_type_class(json_type['type']).normalize_definition(
            json_type)
        if out == '*':
            dim_str = ''
            if json_type['type'] == '1darray':
                if 'length' in json_type:
                    dim_str = ', dimension(%s)' % str(json_type['length'])
            elif json_type['type'] == 'ndarray':
                if 'shape' in json_type:
                    dim_str = ', dimension(%s)' % ','.join(
                        [str(x) for x in json_type['shape']])
            json_subtype = copy.deepcopy(json_type)
            json_subtype['type'] = json_subtype.pop('subtype')
            out = cls.get_native_type(datatype=json_subtype) + dim_str
            if not dim_str:
                json_subtype['type'] = out.split('(')[0]
                if json_subtype['type'] == 'character':  # pragma: debug
                    json_subtype['precision'] = ''
                    raise RuntimeError("Character array requires precision.")
                else:
                    json_subtype['precision'] = int(json_subtype['precision'] / 8)
                json_subtype.setdefault('ndim', 'n')
                out = 'type(%s)' % cls.get_native_type(
                    type=('%s_pointer' % json_type['type'])).format(
                        **json_subtype)
        elif 'X' in out:
            if cls.allows_realloc(kwargs):
                out = 'type(yggchar_r)'
            else:
                if out.startswith('ygguint'):
                    out = 'type(%s)' % out
                if out.startswith('logical'):
                    precision = json_type.get('precision', 8)
                elif out.startswith('complex'):
                    precision = json_type['precision'] / 2
                elif json_type.get('subtype', json_type['type']) == 'unicode':
                    precision = json_type['precision'] / 4
                else:
                    precision = json_type['precision']
                out = out.replace('X', str(int(precision / 8)))
        return out
        
    @classmethod
    def get_json_type(cls, native_type):
        r"""Get the JSON type from the native language type.

        Args:
            native_type (str): The native language type.

        Returns:
            str, dict: The JSON type.

        """
        out = {}
        regex_var = cls.function_param['type_regex']
        grp = re.fullmatch(regex_var, native_type).groupdict()
        if grp['type'].endswith(('_1d', '_nd')):
            regex_nd = r'(?P<type>.*?)(?P<precision>\d+)?_(?P<ndim>(?:1)|(?:n))d'
            grp = re.fullmatch(regex_nd, grp['type']).groupdict()
            if grp['ndim'] == '1':
                grp['shape'] = ':'
            else:
                grp['shape'] = ':,:'
        if grp['type'].startswith('ygguint'):
            grp = {'type': 'ygguintX',
                   'precision': grp['type'].split('ygguint')[-1]}
        if grp.get('precision', False):
            out['precision'] = 8 * int(grp['precision'])
            if grp['type'] == 'complex':
                out['precision'] *= 2
        if (((grp.get('precision', False) or (grp['type'] == 'logical'))
             and (grp['type'] != 'ygguintX'))):
            grp['type'] += '(kind = X)'
        if grp['type'] == 'character':
            if grp.get('length', None):
                out['precision'] = int(grp.get('length', 0)) * 8
            if grp.get('precision_var', None) == 'selected_char_kind(\'ISO_10646\')':
                grp['type'] += '(kind = selected_char_kind(\'ISO_10646\'), len = X)'
                if grp.get('length', None):
                    out['precision'] *= 4
            else:
                grp['type'] += '(len = X)'
        try:
            out['type'] = super(FortranModelDriver, cls).get_json_type(grp['type'])
        except KeyError as e:
            try:
                out['type'] = super(FortranModelDriver, cls).get_json_type(
                    grp['type'] + '(kind = X)')
            except KeyError:  # pragma: debug
                raise e
        if grp.get('shape_var', False):
            if not grp.get('shape', False):
                grp['shape'] = grp['shape_var']
        if grp.get('shape', False):
            shape = grp['shape'].split(',')
            ndim = len(shape)
            out['subtype'] = out['type']
            if ndim == 1:
                out['type'] = '1darray'
                if shape[0] not in '*:':
                    out['length'] = int(shape[0])
            else:
                out['type'] = 'ndarray'
                if shape[0] not in '*:':
                    out['shape'] = [int(i) for i in shape]
        if out['type'] in _valid_types:
            out['subtype'] = out['type']
            out['type'] = 'scalar'
        return out

    @classmethod
    def parse_function_definition(cls, model_file, model_function, **kwargs):
        r"""Get information about the inputs & outputs to a model from its
        defintition if possible.

        Args:
            model_file (str): Full path to the file containing the model
                function's declaration.
            model_function (str): Name of the model function.
            **kwargs: Additional keyword arugments are passed to the parent
                class's method.

        Returns:
            dict: Parameters extracted from the function definitions.

        """
        out = super(FortranModelDriver, cls).parse_function_definition(
            model_file, model_function, **kwargs)
        var_type_map = {}
        for x in re.finditer(cls.function_param['definition_regex'],
                             out['definitions']):
            x_vars = [v.strip() for v in
                      x.groupdict()['name'].split(',')]
            for v in x_vars:
                var_type_map[v] = x.groupdict()['type']
        for x in out.get('inputs', []) + out.get('outputs', []):
            x['native_type'] = var_type_map[x['name']].strip()
            x['datatype'] = cls.get_json_type(x['native_type'])
        return out

    @classmethod
    def update_io_from_function(cls, model_file, model_function, **kwargs):
        r"""Update inputs/outputs from the function definition.

        Args:
            model_file (str): Full path to the file containing the model
                function's declaration.
            model_function (str): Name of the model function.
            **kwargs: Additional keyword arguments are passed to the
                parent class's method.

        Returns:
            dict, None: Flag variable used by the model. If None, the
                model does not use a flag variable.

        """
        out = super(FortranModelDriver, cls).update_io_from_function(
            model_file, model_function, **kwargs)
        for x in kwargs.get('inputs', []) + kwargs.get('outputs', []):
            if x['datatype']['type'] == 'array':
                nvars_items = len(x['datatype'].get('items', []))
                nvars = sum([(not ix.get('is_length_var', False))
                             for ix in x['vars']])
                if nvars_items == nvars:
                    x['use_generic'] = False
                else:
                    x['use_generic'] = True
        return out
        
    @classmethod
    def prepare_variables(cls, vars_list, for_yggdrasil=False, **kwargs):
        r"""Concatenate a set of input variables such that it can be passed as a
        single string to the function_call parameter.

        Args:
            vars_list (list): List of variable dictionaries containing info
                (e.g. names) that should be used to prepare a string representing
                input/output to/from a function call.
            for_yggdrasil (bool, optional): If True, the variables will be
                prepared in the formated expected by calls to yggdarsil
                send/recv methods. Defaults to False.
            **kwargs: Additional keyword arguments are passed to the
                parent class's method.

        Returns:
            str: Concatentated variables list.

        """
        if not isinstance(vars_list, list):
            vars_list = [vars_list]
        new_vars_list = vars_list
        if for_yggdrasil:
            new_vars_list = []
            for v in vars_list:
                assert(isinstance(v, dict))
                # if isinstance(v, dict):
                v = dict(v, name=('yggarg(%s)' % v['name']))
                # else:
                #     v = 'yggarg(%s)' % v
                new_vars_list.append(v)
        out = super(FortranModelDriver, cls).prepare_variables(
            new_vars_list, for_yggdrasil=for_yggdrasil, **kwargs)
        if for_yggdrasil and (len(new_vars_list) > 1):
            return '[%s]' % out
        return out
        
    @classmethod
    def write_executable_import(cls, module=False, **kwargs):
        r"""Add import statements to executable lines.
       
        Args:
            module (str, optional): If provided, the include statement
                importing code will be wrapped in a module of the provided
                name. Defaults to False and the include will not be
                wrapped.
                
            **kwargs: Keyword arguments for import statement.

        Returns:
            list: Lines required to complete the import.
 
        """
        if 'filename' in kwargs:
            if (not module) and os.path.isfile(kwargs['filename']):
                with open(kwargs['filename'], 'r') as fd:
                    contents = fd.read()
                    if 'contains' not in contents.lower():  # pragma: debug
                        raise ValueError("Could not locate 'contains' keyword "
                                         "in user defined module.")
                    idx = contents.lower().index('contains')
                    return (contents[:idx]
                            + '  use %s\n' % cls.interface_library
                            + cls.global_scope_macro + '\n'
                            + contents[idx:]).splitlines()
            kwargs['filename'] = os.path.basename(kwargs['filename'])
        out = super(FortranModelDriver, cls).write_executable_import(
            **kwargs)
        if module:
            out = (['module %s' % module,
                    '  use %s' % cls.interface_library,
                    cls.global_scope_macro,
                    'contains']
                   # Use output directly when import uses directive #include
                   + out
                   # Indent when import uses include statement
                   # + [cls.function_param['indent'] + x for x in out]
                   + ['end module %s' % module])
        return out
        
    @classmethod
    def write_executable(cls, lines, **kwargs):
        r"""Return the lines required to complete a program that will run
        the provided lines.

        Args:
            lines (list): Lines of code to be wrapped as an executable.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            lines: Lines of code wrapping the provided lines with the
                necessary code to run it as an executable (e.g. C/C++'s main).

        """
        if not isinstance(lines, list):
            lines = [lines]
        last_use = 0
        for i, line in enumerate(lines):
            if not line.lstrip().lower().startswith('use'):
                last_use = i - 1
                break
        imports = kwargs.pop('imports', None)
        if imports is not None:
            if not isinstance(imports, list):
                imports = [imports]
            for kws in imports:
                if ('filename' in kws) and os.path.isfile(kws['filename']):
                    with open(kws['filename'], 'r') as fd:
                        contents = fd.read()
                    regex_module = (r'(?i)\s*module\s+(?P<module>.+)\n'
                                    r'(?:.*?\n)*?'
                                    r'\s*end\s+module\s+(?P=module)')
                    match_module = re.search(regex_module, contents)
                    if match_module:
                        module = match_module.groupdict()['module']
                    else:
                        module = '%s_module' % kws['function']
                        kws['module'] = module
                    lines.insert(last_use + 1, 'use %s' % module)
                    last_use += 1
        if 'implicit none' not in lines:
            lines.insert(last_use + 1, 'implicit none')
        out = super(FortranModelDriver, cls).write_executable(
            lines, imports=imports, **kwargs)
        return out
        
    @classmethod
    def write_function_def(cls, function_name, **kwargs):
        r"""Write a function definition.

        Args:
            function_name (str): Name fo the function being defined.
            **kwargs: Additional keyword arguments are passed to the
                parent class's method.

        Returns:
            list: Lines completing the function call.

        """
        if (((not kwargs.get('outputs_in_inputs', False))
             and (len(kwargs.get('outputs', [])) == 0))):
            kwargs.setdefault(
                'function_keys',
                ('subroutine_def_begin', 'subroutine_def_end'))
        # Package is required for new datatypes
        kwargs['skip_interface'] = False
        return super(FortranModelDriver, cls).write_function_def(
            function_name, **kwargs)

    @classmethod
    def escape_quotes(cls, x):
        r"""Escape quotes in a string.

        Args:
           x (str): String to escape quotes in.

        Returns:
           str: x with escaped quotes.

        """
        out = x.replace('"', '""')
        out = out.replace("'", "''")
        return out

    @classmethod
    def split_line(cls, line, length=None, force_split=False):
        r"""Split a line as close to (or before) a given character as
        possible.

        Args:
            line (str): Line to split.
            length (int, optional): Maximum length of split lines. Defaults
                to cls.max_line_width if not provided.
            force_split (bool, optional): If True, force a split to
                occur at the specified length. Defauts to False.

        Returns:
            list: Set of lines resulting from spliting the provided line.

        """
        if line.startswith(('#if', '#endif', '#define', '#else', '#ifdef',
                            '#ifndef')):
            return [line]
        # if line.lstrip().lower().startswith("include"):
        #     force_split = True
        return super(FortranModelDriver, cls).split_line(
            line, length=length, force_split=force_split)

    @classmethod
    def allows_realloc(cls, var, from_native_type=False):
        r"""Determine if a variable allows the receive call to perform
        realloc.

        Args:
            var (dict): Dictionary of variable properties.
            from_native_type (bool, optional): If True, the reallocability
                of the variable will be determined from the native type.
                Defaults to False.

        Returns:
            bool: True if the variable allows realloc, False otherwise.

        """
        if isinstance(var, dict):
            if from_native_type and ('native_type' in var):
                regex_native = r'type\((?:yggchar_r)|(?:.+?\d*_(?:(?:1)|(?:n))d)\)'
                match = re.search(regex_native, var['native_type'])
                if match:
                    return True
            else:
                datatype = var.get('datatype', var)
                if isinstance(datatype, str):
                    datatype = {'type': datatype}
                if (((datatype.get('subtype', datatype.get('type', None))
                      in ['bytes', 'unicode'])
                     and ('precision' not in datatype))):
                    return True
                elif (((datatype.get('type', None) == '1darray')
                       and ('length' not in datatype))):
                    return True
                elif (((datatype.get('type', None) == 'ndarray')
                       and ('shape' not in datatype))):
                    return True
        return False
        
    @classmethod
    def write_declaration(cls, var, **kwargs):
        r"""Return the lines required to declare a variable with a certain
        type.

        Args:
            var (dict, str): Name or information dictionary for the variable
                being declared.
            **kwargs: Additional keyword arguments are passed to the
                parent class's method.

        Returns:
            list: The lines declaring the variable.

        """
        out = super(FortranModelDriver, cls).write_declaration(var, **kwargs)
        if ((cls.allows_realloc(var)
             and (not cls.allows_realloc(var, from_native_type=True)))):
            var = dict(var, name='%s_realloc' % var['name'])
            var.pop('native_type', None)
            out += super(FortranModelDriver, cls).write_declaration(var, **kwargs)
        return out
            
    @classmethod
    def write_model_recv(cls, channel, recv_var, **kwargs):
        r"""Write a model receive call include checking the return flag.

        Args:
            channel (str): Name of variable that the channel being received from
                was stored in.
            recv_var (dict, list): Information of one or more variables that
                receieved information should be stored in.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            list: Lines required to carry out a receive call in this language.

        """
        recv_var_str = recv_var
        out_after = []
        if not isinstance(recv_var, str):
            recv_var_par = cls.channels2vars(recv_var)
            allows_realloc = [cls.allows_realloc(v)
                              for v in recv_var_par]
            if any(allows_realloc):
                kwargs['alt_recv_function'] = cls.function_param['recv_heap']
                new_recv_var_par = []
                for i, v in enumerate(recv_var_par):
                    if (((not cls.allows_realloc(v, from_native_type=True))
                         and allows_realloc[i])):
                        out_after.append('call yggassign(%s_realloc, %s)'
                                         % (v['name'], v['name']))
                        v = dict(v, name=('%s_realloc' % v['name']))
                    new_recv_var_par.append(v)
                recv_var_par = new_recv_var_par
            else:
                kwargs['alt_recv_function'] = cls.function_param['recv_stack']
            recv_var_str = cls.prepare_output_variables(
                recv_var_par, in_inputs=cls.outputs_in_inputs,
                for_yggdrasil=True)
        out = super(FortranModelDriver, cls).write_model_recv(
            channel, recv_var_str, **kwargs)
        return out + out_after

    @classmethod
    def write_model_send(cls, channel, send_var, **kwargs):
        r"""Write a model send call include checking the return flag.

        Args:
            channel (str): Name of variable that the channel being sent to
                was stored in.
            send_var (dict, list): Information on one or more variables
                containing information that will be sent.
            flag_var (str, optional): Name of flag variable that the flag should
                be stored in. Defaults to 'flag',
            allow_failure (bool, optional): If True, the returned lines will
                call a break if the flag is False. Otherwise, the returned
                lines will issue an error. Defaults to False.

        Returns:
            list: Lines required to carry out a send call in this language.

        """
        send_var_str = send_var
        out_before = []
        if not isinstance(send_var, str):
            send_var_par = []
            for v in cls.channels2vars(send_var):
                if (((not cls.allows_realloc(v, from_native_type=True))
                     and cls.allows_realloc(v))):
                    if v.get('datatype', v).get('type', False) in ['1darray']:
                        out_before.append('call yggassign(%s, %s_realloc)'
                                          % (v['name'], v['name']))
                        v = dict(v, name=('%s_realloc' % v['name']))
                send_var_par.append(v)
            send_var_str = cls.prepare_input_variables(
                send_var_par, for_yggdrasil=True)
        out = super(FortranModelDriver, cls).write_model_send(
            channel, send_var_str, **kwargs)
        return out_before + out
    
    @classmethod
    def write_print_var(cls, var, **kwargs):
        r"""Get the lines necessary to print a variable in this language.

        Args:
            var (dict): Variable information.
            **kwargs: Additional keyword arguments are passed to the
                parent class's method.

        Returns:
            list: Lines printing the specified variable.

        """
        if isinstance(var, dict):
            datatype = var.get('datatype', var)
            typename = datatype.get('type', None)
            if ((((typename == '1darray') and ('length' not in datatype))
                 or ((typename == 'ndarray') and ('shape' not in datatype)))):
                return []
        return super(FortranModelDriver, cls).write_print_var(
            var, **kwargs)

    @classmethod
    def write_type_def(cls, name, datatype, **kwargs):
        r"""Get lines declaring the data type within the language.

        Args:
            name (str): Name of variable that definition should be stored in.
            datatype (dict): Type definition.
            **kwargs: Additional keyword arguments are passed to the
                parent class's method.

        Returns:
            list: Lines required to define a type definition.

        """
        if datatype['type'] == 'array':
            datatype.setdefault('items', [])
        elif datatype['type'] == 'object':
            datatype.setdefault('properties', {})
        elif datatype['type'] == 'ndarray':
            if 'shape' not in datatype:
                datatype = dict(datatype, shape=[])
            if 'ndim' not in datatype:
                datatype = dict(datatype, ndim=len(datatype['shape']))
        if datatype.get('subtype', datatype['type']) in ['bytes', 'unicode']:
            if 'precision' not in datatype:
                datatype = dict(datatype, precision=0)
        elif datatype.get('subtype', datatype['type']) in _valid_types:
            datatype.setdefault('precision', 32)
        out = super(FortranModelDriver, cls).write_type_def(
            name, datatype, **kwargs)
        return out
    
    @classmethod
    def write_type_decl(cls, name, datatype, **kwargs):
        r"""Get lines declaring the datatype within the language.

        Args:
            name (str): Name of variable that should be declared.
            datatype (dict): Type definition.
            **kwargs: Additional keyword arguments are passed to the
                parent class's method.

        Returns:
            list: Lines required to define a type declaration.

        """
        if datatype['type'] == 'array':
            datatype.setdefault('items', [])
        elif datatype['type'] == 'object':
            datatype.setdefault('properties', {})
        elif datatype['type'] == 'ndarray':
            if 'shape' not in datatype:
                datatype = dict(datatype, shape=[])
            if 'ndim' not in datatype:
                datatype = dict(datatype, ndim=len(datatype['shape']))
        out = super(FortranModelDriver, cls).write_type_decl(
            name, datatype, **kwargs)
        return out



import numpy as np
from yggdrasil import units
from yggdrasil.tools import safe_eval
from yggdrasil.communication.filters.FilterBase import FilterBase


class StatementFilter(FilterBase):
    r"""Class for filtering messages based on a provided statement using Python syntax.

    Args:
        statement (str): Python statement in terms of the message as represented by
            the string "%x%" that should evaluate to a boolean, True if the message
            should pass through the filter, False if it should not. The statement
            should only use a limited set of builtins and the math library (See
            yggdrasil.tools.safe_eval). If more complex relationships are required,
            use the FunctionFilter class.

    Attributes:
        statement (str): Python statement that will be evaluated to determine if
            messages should or should not pass the filter.

    """
    _filtertype = 'statement'
    _schema_required = ['statement']
    _schema_properties = {'statement': {'type': 'string'}}

    def __init__(self, *args, **kwargs):
        super(StatementFilter, self).__init__(*args, **kwargs)
        self.statement = self.statement.replace('%x%', 'x')

    def evaluate_filter(self, x):
        r"""Call filter on the provided message.

        Args:
            x (object): Message object to filter.

        Returns:
            bool: True if the message will pass through the filter, False otherwise.

        """
        return safe_eval(self.statement, x=x)

    @classmethod
    def get_testing_options(cls):
        r"""Get testing options for the filter class.

        Returns:
            list: Mutiple dictionaries of keywords and messages that will
                pass/fail for those keywords.
        
        """
        out = [{'kwargs': {'statement': '%x% != 2'},
                'pass': [1, 3], 'fail': [2]},
               {'kwargs': {'statement': '%x% != array([0, 0, 0])'},
                'pass': [np.ones(3, int)], 'fail': [np.zeros(3, int)]},
               {'kwargs': {'statement': '%x% != add_units(1, "cm")'},
                'pass': [units.add_units(2, 'cm')],
                'fail': [units.add_units(1, 'cm')]},
               {'kwargs': {'statement': '%x% != '
                           + repr(units.add_units(1, 'cm'))},
                'pass': [units.add_units(2, 'cm')],
                'fail': [units.add_units(1, 'cm')]}]
        return out



from yggdrasil.examples.tests import ExampleTstBase


class TestExampleTimesync2(ExampleTstBase):
    r"""Test the timesync2 example."""

    example_name = 'timesync2'
    env = {'TIMESYNC_TSTEP_A': '20', 'TIMESYNC_TSTEP_B': '3'}



# -*- coding: utf-8 -*-
#
# yggdrasil documentation build configuration file, created by
# sphinx-quickstart on Wed Sep  6 12:03:29 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import glob
import sys
import sphinx_rtd_theme
import yggdrasil
from yggdrasil import tools, languages, components
# sys.path.insert(0, os.path.abspath('.'))
doxydir = os.path.join(os.path.abspath('../'), "doxy", "xml")
rootdir = os.path.abspath('../../')
srcdir = os.path.join(rootdir, "yggdrasil")
sys.path.append(doxydir)


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon',
    'breathe',
]


relative_files = []
for lang in tools.get_supported_lang():
    if lang == 'python':
        continue
    try:
        lang_dir = languages.get_language_dir(lang)
        lang_ext = components.import_component('model', lang).get_language_ext()
    except ValueError:
        continue
    for iext in lang_ext:
        relative_files += [os.path.relpath(f, start=srcdir) for f in
                           glob.glob(os.path.join(lang_dir, '*' + iext))]
        relative_files += [os.path.relpath(f, start=srcdir) for f in
                           glob.glob(os.path.join(lang_dir, '*', '*' + iext))]

breathe_projects = {"yggdrasil": doxydir}
breathe_default_project = "yggdrasil"
breathe_projects_source = {"yggdrasil": (srcdir, relative_files)}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'yggdrasil'
copyright = u'2017, Meagan Lang, David Raila'
author = u'Meagan Lang, David Raila'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = yggdrasil.__version__
# The short X.Y version.
version = release.split('+')[0]

# Substitutions
# .. _Docs: http://yggdrasil.readthedocs.io/en/latest/
rst_epilog = """
.. _Docs: https://cropsinsilico.github.io/yggdrasil/
.. _Meagan Lang: langmm.astro@gmail.com
.. |yggdrasil| replace:: yggdrasil
"""

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_context = {
    'css_files': [
        '_static/theme_overrides.css',  # override wide tables in RTD theme
        ],
     }


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'yggdrasildoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'yggdrasil.tex', u'ygg\\_interface Documentation',
     u'Meagan Lang, David Raila', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'yggdrasil', u'yggdrasil Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'yggdrasil', u'yggdrasil Documentation',
     author, 'yggdrasil', 'One line description of project.',
     'Miscellaneous'),
]



# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']



# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None}



import sys
import numpy as np
from yggdrasil import units
from yggdrasil.interface.YggInterface import (
    YggTimesync, YggOutput)


def xvar2x(xvar):
    r"""Convert xvar to x."""
    x = 2 * xvar
    return x


def x2xvar(x):
    r"""Convert x to xvar."""
    xvar = x / 2
    return xvar


def xagg(series):
    r"""Aggregate x variables."""
    return max(series, key=abs)


def merge_z(z1, z2):
    r"""Merge the z1 and z2 variables."""
    return z1 + z2


def split_z(z):
    r"""Split z into z1 and z2 variables."""
    return (z / 2.0, z / 2.0)


def timestep_calc(t, model):
    r"""Updates the state based on the time where x is a sine wave
    with period of 10 days and y is a cosine wave with a period of 5 days.
    If model is 'A', the forth state variable will be 'a', a sine
    with a period of 2.5 days. If model is 'B', the forth state
    variable will be 'b', a cosine with a period of 2.5 days.

    Args:
        t (float): Current time.
        model (str): Identifier for the model (A or B).

    Returns:
        dict: Map of state parameters.

    """
    if model == 'A':
        state = {
            'x': np.sin(2.0 * np.pi * t / units.add_units(10, 'day')),
            'y': np.cos(2.0 * np.pi * t / units.add_units(5, 'day')),
            'z1': -np.cos(2.0 * np.pi * t / units.add_units(20, 'day')),
            'z2': -np.cos(2.0 * np.pi * t / units.add_units(20, 'day')),
            'a': np.sin(2.0 * np.pi * t / units.add_units(2.5, 'day'))}
    else:
        state = {
            'xvar': x2xvar(np.sin(2.0 * np.pi * t / units.add_units(10, 'day'))),
            'yvar': np.cos(2.0 * np.pi * t / units.add_units(5, 'day')),
            'z': -2.0 * np.cos(2.0 * np.pi * t / units.add_units(20, 'day')),
            'b': np.cos(2.0 * np.pi * t / units.add_units(2.5, 'day'))}
    return state


def main(t_step, t_units, model):
    r"""Function to execute integration.

    Args:
        t_step (float): The time step that should be used.
        t_units (str): Units of the time step.
        model (str): Identifier for the model (A or B).

    """
    print('Hello from Python timesync: timestep = %s %s' % (t_step, t_units))
    t_step = units.add_units(t_step, t_units)
    t_start = units.add_units(0.0, t_units)
    t_end = units.add_units(5.0, 'day')
    state = timestep_calc(t_start, model)

    # Set up connections matching yaml
    # Timestep synchronization connection will be 'statesync'
    timesync = YggTimesync("statesync")
    out = YggOutput('output')

    # Initialize state and synchronize with other models
    t = t_start
    ret, state = timesync.call(t, state)
    if not ret:
        raise RuntimeError("timesync(Python): Initial sync failed.")
    print('timesync(Python): t = % 8s' % t, end='')
    for k, v in state.items():
        print(', %s = %+ 5.2f' % (k, v), end='')
    print('')

    # Send initial state to output
    flag = out.send(dict(state, time=t))
    if not flag:
        raise RuntimeError("timesync(Python): Failed to send "
                           "initial output for t=%s." % t)
    
    # Iterate until end
    while t < t_end:

        # Perform calculations to update the state
        t = t + t_step
        state = timestep_calc(t, model)

        # Synchronize the state
        ret, state = timesync.call(t, state)
        if not ret:
            raise RuntimeError("timesync(Python): sync for t=%f failed." % t)
        print('timesync(Python): t = % 8s' % t, end='')
        for k, v in state.items():
            print(', %s = %+ 5.2f' % (k, v), end='')
        print('')

        # Send output
        flag = out.send(dict(state, time=t))
        if not flag:
            raise RuntimeError("timesync(Python): Failed to send output for t=%s." % t)

    print('Goodbye from Python timesync')


if __name__ == '__main__':
    # Take time step from the first argument
    main(float(sys.argv[1]), sys.argv[2], sys.argv[3])



import os
from yggdrasil.examples.tests import ExampleTstBase


class TestExampleWofost(ExampleTstBase):
    r"""Test the Wofost example."""

    example_name = 'wofost'

    @property
    def output_files(self):
        r"""list: Output files for the run."""
        if self.yamldir is None:  # pragma: debug
            return None
        return [os.path.join(self.yamldir, 'output.txt')]

    def check_results(self):
        r"""This should be overridden with checks for the result."""
        pass



from yggdrasil import units


def modelC_function(in_val):
    out_val = 2 * in_val
    print("modelC_function(%s) = %s" % (in_val, out_val))
    return in_val, out_val


def transform_function(in_val):
    return (in_val / units.add_units(10, 'cm**2'))



import os
from yggdrasil.examples.tests import ExampleTstBase


class TestExampleTransformedIO(ExampleTstBase):
    r"""Test the transformed_io example."""

    example_name = 'transformed_io'

    @property
    def expected_output_files(self):
        r"""list: Examples of expected output for the run."""
        return [os.path.join(self.yamldir, 'Output', 'outputB.txt'),
                os.path.join(self.yamldir, 'Output', 'outputC.txt')]
        
    @property
    def output_files(self):
        r"""list: Output files for the run."""
        return [os.path.join(self.yamldir, 'outputB.txt'),
                os.path.join(self.yamldir, 'outputC.txt')]



"""Constants used by yggdrasil."""


LANG2EXT = {'python': '.py',
            'matlab': '.m',
            'r': '.R',
            'c': '.c',
            'cpp': '.cpp',
            'c++': '.cpp',
            'fortran': '.f90',
            'yaml': '.yml'}
EXT2LANG = {v: k for k, v in LANG2EXT.items()}



import os
import numpy as np
import unittest
from yggdrasil import units
from yggdrasil.tests import assert_equal
from yggdrasil.communication import AsciiTableComm
from yggdrasil.communication.tests import test_AsciiFileComm as parent
from yggdrasil.metaschema.properties.ScalarMetaschemaProperties import (
    data2dtype)


def test_AsciiTableComm_nofmt():
    r"""Test read of asciitable without format."""
    test_file = os.path.join(os.getcwd(), 'temp_file.txt')
    rows = [('one', 1, 1.0), ('two', 2, 2.0), ('three', 3, 3.0)]
    lines = [('%5s\t%d\t%f\n' % r) for r in rows]
    contents = (''.join(lines)).encode("utf-8")
    with open(test_file, 'wb') as fd:
        fd.write(contents)
    inst = AsciiTableComm.AsciiTableComm('test', test_file, direction='recv')
    inst.open()
    for ans in rows:
        flag, x = inst.recv_dict()
        assert(flag)
        irow = [e for e in ans]
        irow[0] = irow[0].encode("utf-8")
        idict = {'f%d' % i: irow[i] for i in range(len(irow))}
        # irow = tuple(irow)
        assert_equal(x, idict)
    flag, x = inst.recv()
    assert(not flag)
    inst.close()
    os.remove(test_file)


class TestAsciiTableComm(parent.TestAsciiFileComm):
    r"""Test for AsciiTableComm communication class."""

    comm = 'AsciiTableComm'
    
    @unittest.skipIf(True, 'Table comm')
    def test_send_recv_comment(self):
        r"""Disabled: Test send/recv with commented message."""
        pass  # pragma: no cover

    def map_sent2recv(self, obj):
        r"""Convert a sent object into a received one."""
        if not self.instance.is_eof(obj):
            field_units = self.testing_options.get('field_units', None)
            if field_units:
                if isinstance(obj, dict):
                    return {k: units.add_units(v, u, dtype=data2dtype(v))
                            for (k, v), u in zip(obj.items(), field_units)}
                elif isinstance(obj, (list, tuple)):
                    return [units.add_units(x, u, dtype=data2dtype(x))
                            for x, u in zip(obj, field_units)]
        return obj

    
class TestAsciiTableComm_AsArray(TestAsciiTableComm):
    r"""Test for AsciiTableComm communication class."""

    testing_option_kws = {'array_columns': True}


class TestAsciiTableComm_single(TestAsciiTableComm):
    r"""Test for AsciiTableComm communication class with field names sent."""

    def get_options(self):
        r"""Get testing options."""
        nele = 5
        dtype = np.dtype(dict(formats=['float'], names=['f0']))
        arr1 = np.zeros((nele, ), dtype)
        arr2 = np.ones((nele, ), dtype)
        out = {'kwargs': {'as_array': True, 'field_names': ['f0']},
               'contents': (
                   b'# f0\n# %g\n'
                   + nele * b'0\n' + nele * b'1\n'),
               'send': [[arr1['f0']], [arr2['f0']]],
               'recv': [[np.hstack([arr1, arr2])['f0']]],
               'recv_partial': [[[arr1['f0']]], [[arr2['f0']]]],
               'dict': {'f0': arr1['f0']},
               'objects': [[arr1['f0']], [arr2['f0']]]}
        out['msg'] = out['send'][0]
        out['msg_array'] = arr1
        return out

    def test_send_dict_default(self):
        r"""Test automated conversion of dictionary to pandas data frame."""
        self.do_send_recv(msg_send=self.testing_options['dict'],
                          msg_recv=self.testing_options['msg'])



import os
import sys
PY_MAJOR_VERSION = sys.version_info[0]
IS_WINDOWS = (sys.platform in ['win32', 'cygwin'])
# Import config parser
try:
    from ConfigParser import RawConfigParser as HandyConfigParser
except ImportError:
    try:
        from configparser import RawConfigParser as HandyConfigParser
    except ImportError:
        HandyConfigParser = None


def add_excl_rule(excl_list, new_rule):
    r"""Add an exclusion rule to the list if its not in there.

    Args:
        excl_list (list): List of exclusion rules to update.
        new_rule (str): New rule to add to the list if it already exists.

    Returns:
        list: Updated exclustion rules.

    """
    if new_rule not in excl_list:
        excl_list.append(new_rule)
    return excl_list


def rm_excl_rule(excl_list, new_rule):
    r"""Remove an exclusion rule from the list if its in there.

    Args:
        excl_list (list): List of exclusion rules to update.
        new_rule (str): New rule to remove from the list if it exists.

    Returns:
        list: Updated exclustion rules.

    """
    if new_rule in excl_list:
        excl_list.remove(new_rule)
    return excl_list


def create_coveragerc(installed_languages):
    r"""Create the coveragerc to reflect the OS, Python version, and availability
    of matlab. Parameters from the setup.cfg file will be added. If the
    .coveragerc file already exists, it will be read first before adding setup.cfg
    options.
    

    Args:
        installed_languages (dict): Dictionary of language/boolean key/value
            pairs indicating optional languages and their state of installation.

    Returns:
        bool: True if the file was created/updated successfully, False otherwise.

    """
    if HandyConfigParser is None:
        return False
    debug_msg = 'cwd = %s, os.path.dirname(__file__) = %s' % (
        os.getcwd(), os.path.dirname(__file__))
    print(debug_msg)
    # covdir = os.path.dirname(__file__)
    covdir = os.getcwd()
    covrc = os.path.join(covdir, '.coveragerc')
    cp = HandyConfigParser("")
    # Read from existing .coveragerc
    if os.path.isfile(covrc):
        cp.read(covrc)
    # Read options from setup.cfg
    setup_cfg = os.path.join(os.path.dirname(__file__), 'setup.cfg')
    cp_cfg = HandyConfigParser("")
    cp_cfg.read(setup_cfg)
    # Transfer options
    for x in cp_cfg.sections():
        if x.startswith('coverage:'):
            sect_cp = x.split('coverage:')[-1]
            if not cp.has_section(sect_cp):
                cp.add_section(sect_cp)
            for opt in cp_cfg.options(x):
                if cp.has_option(sect_cp, opt):
                    val_old = [l.strip() for l in cp.get(sect_cp, opt).split('\n')]
                    val_new = [l.strip() for l in cp_cfg.get(x, opt).split('\n')]
                    for v in val_new:
                        if v not in val_old:
                            val_old.append(v)
                    opt_new = '\n'.join(val_old)
                else:
                    opt_new = cp_cfg.get(x, opt)
                cp.set(sect_cp, opt, opt_new)
    # Exclude rules for all files
    if not cp.has_section('report'):
        cp.add_section('report')
    if cp.has_option('report', 'exclude_lines'):
        excl_str = cp.get('report', 'exclude_lines')
        excl_list = excl_str.strip().split('\n')
    else:
        excl_list = []
    # Platform
    if IS_WINDOWS:
        excl_list = rm_excl_rule(excl_list, 'pragma: windows')
    else:
        excl_list = add_excl_rule(excl_list, 'pragma: windows')
    # Python version
    verlist = [2, 3]
    for v in verlist:
        vincl = 'pragma: Python %d' % v
        if PY_MAJOR_VERSION == v:
            excl_list = rm_excl_rule(excl_list, vincl)
        else:
            excl_list = add_excl_rule(excl_list, vincl)
    # Language specific
    for k, v in installed_languages.items():
        if v:
            excl_list = add_excl_rule(excl_list, 'pragma: no %s' % k)
            excl_list = rm_excl_rule(excl_list, 'pragma: %s' % k)
        else:
            excl_list = add_excl_rule(excl_list, 'pragma: %s' % k)
            excl_list = rm_excl_rule(excl_list, 'pragma: no %s' % k)
    # Add new rules
    cp.set('report', 'exclude_lines', '\n' + '\n'.join(excl_list))
    # Write
    with open(covrc, 'w') as fd:
        cp.write(fd)
    return True


if __name__ == "__main__":
    LANG_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                             'yggdrasil', 'languages')
    sys.path.insert(0, LANG_PATH)
    try:
        import install_languages
    finally:
        sys.path.pop(0)
    installed_languages = install_languages.install_all_languages()
    flag = create_coveragerc(installed_languages)
    if not flag:
        raise Exception("Failed to create/update converagerc file.")
